----------------------------------------------------------------------------------
-- Engineer:		Armin Diehl
-- Create Date:		20:49:41 02/10/2007
-- Design Name:		ECB Z80 Businterface for VGA Terminal
-- Module Name:		ECB_BusIntf - Behavioral 
-- Project Name:	VGA Term
-- Target Devices:	xc2s200
-- Tool versions:	ise 9.1.01i
-- Description: 
-- Revision		0.01 - File Created
----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

PACKAGE ECB_BusIntf_pckg IS
  COMPONENT ECB_BusIntf IS
  Port ( 
	res_n			: in    STD_LOGIC;
	clk			: in    STD_LOGIC;
	-- ECB Bus
	ECB_clk			: in    STD_LOGIC;
	rd_n			: in    STD_LOGIC;
	wr_n			: in    STD_LOGIC;
	iorq_n			: in    STD_LOGIC;
	A			: in    STD_LOGIC_VECTOR (7 downto 0);
	D			: inout STD_LOGIC_VECTOR (7 downto 0);
	-- Keyboard
	KbdChar_i		: in    STD_LOGIC_VECTOR (7 downto 0);
	KbdCharAvail_i		: in    STD_LOGIC;
	KbdCharProcessed_o	: out   STD_LOGIC;
	-- Terminal Emulation
	TermChar_o		: out   STD_LOGIC_VECTOR (7 downto 0);
	TermCharAvail_o		: out   STD_LOGIC;
	TermReady		: in    STD_LOGIC;
	ps2FrameErr		: in	STD_LOGIC;
	ps2ParityErr		: in	STD_LOGIC;
	ps2LowCharAvail		: in	STD_LOGIC;
	StateHistory		: in STD_LOGIC_VECTOR (63 downto 0);
	LastCharProcessed	: in STD_LOGIC_VECTOR (7 downto 0);
	KeyHistory		: in STD_LOGIC_VECTOR (63 downto 0));
  END COMPONENT;
END PACKAGE;

----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;


-- IOADDR RD	-> Status Port, Bit 7 = KbdChar available, Bit 6 = Ready for output
-- IOADDR+1 RD	-> Read keyboard char
-- IOADDR+1 WR	-> Write terminal char

entity ECB_BusIntf is
    generic(
	IOADDR		: STD_LOGIC_VECTOR (7 downto 0) := x"C0"	-- grip: C0
    );
    Port ( 
	res_n			: in    STD_LOGIC;
	clk			: in    STD_LOGIC;
	-- ECB Bus
	ECB_clk			: in    STD_LOGIC;
	rd_n			: in    STD_LOGIC;
	wr_n			: in    STD_LOGIC;
	iorq_n			: in    STD_LOGIC;
	A			: in    STD_LOGIC_VECTOR (7 downto 0);
	D			: inout STD_LOGIC_VECTOR (7 downto 0);
	-- Keyboard
	KbdChar_i		: in    STD_LOGIC_VECTOR (7 downto 0);
	KbdCharAvail_i		: in    STD_LOGIC;
	KbdCharProcessed_o	: out   STD_LOGIC;
	-- Terminal Emulation
	TermChar_o		: out   STD_LOGIC_VECTOR (7 downto 0);
	TermCharAvail_o		: out   STD_LOGIC;
	TermReady		: in    STD_LOGIC;
	ps2FrameErr		: in	STD_LOGIC;
	ps2ParityErr		: in	STD_LOGIC;
	ps2LowCharAvail		: in	STD_LOGIC;
	StateHistory		: in STD_LOGIC_VECTOR (63 downto 0);
	LastCharProcessed	: in STD_LOGIC_VECTOR (7 downto 0);
	KeyHistory		: in STD_LOGIC_VECTOR (63 downto 0));
end ECB_BusIntf;

architecture Behavioral of ECB_BusIntf is
SIGNAL CS,CS_READ,CS_WRITE		: STD_LOGIC;
SIGNAL CSECB,CSECB_READ,CSECB_WRITE	: STD_LOGIC;
SIGNAL CS_on_LastClock,CS_off_LastClock : STD_LOGIC;
SIGNAL Addr				: STD_LOGIC_VECTOR(3 downto 0);
SIGNAL RegisterNo,Data			: STD_LOGIC_VECTOR(7 downto 0);
begin
 
  -- output data to host on read 
  setDataRd:PROCESS (CSECB_READ,A,KbdCharAvail_i,TermReady,KbdChar_i,ps2LowCharAvail,ps2FrameErr,ps2ParityErr,StateHistory,iorq_n,LastCharProcessed)
  BEGIN
    IF (CSECB_READ = '1') AND (IORQ_n = '0') THEN	-- IORQ to release bus independent of ECB_clk
      CASE A(3 downto 0) IS
	WHEN "0000" 	=> D <= KbdCharAvail_i & TermReady & "000" & ps2LowCharAvail & ps2FrameErr & ps2ParityErr;
        WHEN "0001"	=> D <= KbdChar_i;
	WHEN "0010"	=> D <= LastCharProcessed;
	
	WHEN "0011"	=> CASE RegisterNo IS
			     WHEN x"00" =>	D <= StateHistory(7 downto 0);
			     WHEN x"01" =>	D <= StateHistory(15 downto 8);
			     WHEN x"02" =>	D <= StateHistory(23 downto 16);
			     WHEN x"03"	=> 	D <= StateHistory(31 downto 24);
			     WHEN x"04" =>	D <= StateHistory(39 downto 32);
			     WHEN x"05" =>	D <= StateHistory(47 downto 40);
			     WHEN x"06" =>	D <= StateHistory(55 downto 48);
			     WHEN x"07" =>	D <= StateHistory(63 downto 56);
			     
			     WHEN x"10" =>	D <= KeyHistory(7 downto 0);
			     WHEN x"11" =>	D <= KeyHistory(15 downto 8);
			     WHEN x"12" =>	D <= KeyHistory(23 downto 16);
			     WHEN x"13"	=> 	D <= KeyHistory(31 downto 24);
			     WHEN x"14" =>	D <= KeyHistory(39 downto 32);
			     WHEN x"15" =>	D <= KeyHistory(47 downto 40);
			     WHEN x"16" =>	D <= KeyHistory(55 downto 48);
			     WHEN x"17" =>	D <= KeyHistory(63 downto 56);
			     
			     WHEN Others=>	D <= (Others => '0');
			   END CASE;
	WHEN others	=> D <= (Others => '0');
      END CASE;
    ELSE
      D <= "ZZZZZZZZ";
    END IF;
  END PROCESS;
  
  -- generate CSECB* from ECB-Bus using the clock from the bus
  genCSECB:PROCESS (ECB_clk,res_n)
  BEGIN
    IF res_n = '0' THEN
      CSECB <= '0'; CSECB_READ <= '0'; CSECB_WRITE <= '0';
      TermChar_o <= (Others => '0');
    ELSIF Rising_Edge(ECB_clk) THEN			-- already there on falling_edge
      IF (IORQ_n = '0') AND (A (7 downto 4) = IOADDR (7 downto 4)) AND 
         ((RD_n = '0') OR (WR_n = '0')) THEN
	CSECB <= '1';
	CSECB_READ <= not RD_n;
	CSECB_WRITE <= not WR_n;
	TermChar_o <= D;				-- set data to termemu for host write
	Data <= D;
	Addr <= A(3 downto 0);
      ELSIF IORQ_n = '1' THEN				-- if iorq is gone, clear the stuff
        CSECB <= '0';
	CSECB_READ <= '0';
	CSECB_WRITE <= '0';
      END IF;
    END IF;
  END PROCESS;
  
  
  genCS:PROCESS (clk,res_n)				-- generate CS with system clock
  BEGIN							-- the ECB_CS has to be active for at least 2 clocks
    IF res_n = '0' THEN
      CS_on_LastClock <= '0';
      CS_off_LastClock <= '0';
      CS_READ <= '0';
      CS_WRITE <= '0';
      CS <= '0';
      RegisterNo <= (Others => '0');
    ELSIF Rising_Edge(clk) THEN
      IF CSECB = '1' THEN				-- if selected
        IF CS_on_LastClock = '0' THEN			-- and not already on last clock
	  CS_on_LastClock <= '1';			-- set flag
	ELSE						-- cs was active on last clock
	  CS <= '1'; CS_READ <= CSECB_READ; 		-- set CS*
	  CS_WRITE <= CSECB_WRITE;
	  CS_off_LastClock <= '0';
	END IF;
      ELSE						-- no (longer) selected
        IF CS_off_LastClock = '0' THEN			-- also on last clock
	  CS_off_LastClock <= '1';			-- no, set flag
	ELSE
	  CS_on_LastClock <= '0';
	  CS <= '0'; CS_READ <= '0';			-- select gone
	  CS_WRITE <= '0';
	END IF;
      END IF;
      
      IF (CS_WRITE = '1') AND (Addr(0) = '1') THEN
	TermCharAvail_o <= '1';
      ELSE
        TermCharAvail_o <= '0';
      END IF;

      IF (CS_WRITE = '1') AND (Addr(0) = '0') THEN
	RegisterNo <= Data;
      END IF;

    END IF;
  END PROCESS;
  
  
  genCharRead:PROCESS (clk,res_n)
  BEGIN
    IF res_n = '0' THEN
      KbdCharProcessed_o <= '0';
    ELSIF Rising_Edge(clk) THEN
      IF (CS_READ = '1') AND (Addr(0)='1') THEN
        KbdCharProcessed_o <= '1';
      ELSE
        KbdCharProcessed_o <= '0';	-- char has been read
      END IF;
    END IF;
  END PROCESS;
  
END Behavioral;

