----------------------------------------------------------------------------------
-- Engineer:       Armin Diehl
-- Create Date:    20:29:25 01/04/2007 
-- Modify Date:    02/03/2007 
-- Module Name:    Led7seg - Behavioral 
-- Description:    Driver for 7 Segemt LED on xs200
--                 Supports 4 Bit, 8 Bit and 16 Bit on one 7 Seg Display
----------------------------------------------------------------------------------

-- 7 Segment for one Digit
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;


package led_pckg is
  component Led7seg is
    Port ( Din : in  STD_LOGIC_VECTOR (4 downto 0);
           Dout : out  STD_LOGIC_VECTOR (7 downto 0));  -- Bit7=DP,Bit6=A,5=B,...
  end component;
  
  component Led7seg8 is
    Port ( clk : in STD_LOGIC;
           reset : in STD_LOGIC;
           Din : in  STD_LOGIC_VECTOR (7 downto 0);
           Dout : out  STD_LOGIC_VECTOR (7 downto 0));  -- Bit7=DP,Bit6=A,5=B,...
  end component;
  
  component Led7seg16 is
    Port ( clk : in STD_LOGIC;
           reset : in STD_LOGIC;
           Din : in  STD_LOGIC_VECTOR (15 downto 0);
           Dout : out  STD_LOGIC_VECTOR (7 downto 0));  -- Bit7=DP,Bit6=A,5=B,...
  end component;
  
end package led_pckg;


library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;


entity Led7seg is
    Port ( Din : in  STD_LOGIC_VECTOR (4 downto 0);
           Dout : out  STD_LOGIC_VECTOR (7 downto 0));  -- Bit7=DP,Bit6=A,5=B,...
end Led7seg;

architecture Behavioral of Led7seg is
SIGNAL X:INTEGER RANGE 0 TO 15;
begin
  X <= conv_integer(Din(3 downto 0));
  Dout(7) <= Din(4);
  WITH X SELECT
    Dout(6 downto 0) <=
    -- ABCDEFG
      "1111110" WHEN  0,
      "0110000" WHEN  1,
      "1101101" WHEN  2,
      "1111001" WHEN  3,
      "0110011" WHEN  4,
      "1011011" WHEN  5,
      "1011111" WHEN  6,
      "1110000" WHEN  7,
      "1111111" WHEN  8,
      "1111011" WHEN  9,
      "0011101" WHEN  10,  -- A
      "0011111" WHEN  11,  -- B
      "0001101" WHEN  12,  -- C
      "0111101" WHEN  13,  -- D
      "1001111" WHEN  14,  -- E
      "1000111" WHEN  15;  -- F

end Behavioral;

-------------------------------------------------------------------------------------------

-- two Digits on one 7 seg display, first digit with decimal point
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
--use WORK.Led7Seg.Led7Seg;

entity Led7seg8 is
    Port ( clk : in STD_LOGIC;
           reset : in STD_LOGIC;
           Din : in  STD_LOGIC_VECTOR (7 downto 0);
           Dout : out  STD_LOGIC_VECTOR (7 downto 0));  -- Bit7=DP,Bit6=A,5=B,...
end Led7seg8;

architecture Behavioral of Led7seg8 is
SIGNAL SegmentData : STD_LOGIC_VECTOR (4 downto 0);
SIGNAL counter : unsigned(25 downto 0);
SIGNAL Tick : STD_LOGIC;
SIGNAL SegNo : INTEGER RANGE 0 TO 1;

COMPONENT Led7Seg is
    Port ( Din : in  STD_LOGIC_VECTOR (4 downto 0);
           Dout : out  STD_LOGIC_VECTOR (7 downto 0));
END COMPONENT;

begin
  Segment:Led7Seg PORT MAP(
    Din  => SegmentData,
    Dout => Dout);
    
  cnt:process(reset,clk)
  begin
    if reset = '1' then
      counter <= (others => '0');
    elsif rising_edge(clk) then
      counter <= counter + 1;
    end if;
  end process;
  
  display:PROCESS(Tick,Reset)
  BEGIN
    IF Reset = '1' THEN
      SegNo <= 0;
    ELSIF Rising_Edge(Tick) THEN
      CASE SegNo IS
        WHEN 0 => SegmentData <= "1" & Din(7 downto 4);
             SegNo <= SegNo + 1;
        WHEN 1 => SegmentData <= "0" & Din(3 downto 0);
             SegNo <= 0;
        WHEN others => NULL;
      END CASE;
    END IF;
  END PROCESS;
  
  Tick <= counter(25);

end Behavioral;

-----------------------------------------------------------------------------------------

-- 4 Digits on one 7 seg display, first digit with decimal point
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
--use WORK.Led7Seg.Led7Seg;

entity Led7seg16 is
    Port ( clk : in STD_LOGIC;
           reset : in STD_LOGIC;
           Din : in  STD_LOGIC_VECTOR (15 downto 0);
           Dout : out  STD_LOGIC_VECTOR (7 downto 0));  -- Bit7=DP,Bit6=A,5=B,...
end Led7seg16;

architecture Behavioral of Led7seg16 is
SIGNAL SegmentData : STD_LOGIC_VECTOR (4 downto 0);
SIGNAL counter : unsigned(25 downto 0);
SIGNAL Tick : STD_LOGIC;
SIGNAL SegNo : INTEGER RANGE 0 TO 3;

COMPONENT Led7Seg is
    Port ( Din : in  STD_LOGIC_VECTOR (4 downto 0);
           Dout : out  STD_LOGIC_VECTOR (7 downto 0));
END COMPONENT;

begin
  Segment:Led7Seg PORT MAP(
    Din  => SegmentData,
    Dout => Dout);
    
  cnt:process(clk,Reset)
  begin
    if reset = '1' then
      counter <= (others => '0');
    elsif rising_edge(clk) then
      counter <= counter + 1;
    end if;
  end process;
  
  display:PROCESS(Tick,Reset)
  BEGIN
    IF Reset = '1' THEN
      SegNo <= 0;
    ELSIF Rising_Edge(Tick) THEN
      CASE SegNo IS
        WHEN 0 => SegmentData <= "1" & Din(15 downto 12);
             SegNo <= SegNo + 1;
        WHEN 1 => SegmentData <= "0" & Din(11 downto 8);
             SegNo <= SegNo + 1;
        WHEN 2 => SegmentData <= "0" & Din(7 downto 4);
             SegNo <= SegNo + 1;
        WHEN 3 => SegmentData <= "0" & Din(3 downto 0);
             SegNo <= 0;
        WHEN others => NULL;
      END CASE;
    END IF;
  END PROCESS;
  
  Tick <= counter(25);
end Behavioral;
