----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    15:33:51 01/03/2007 
-- Design Name: 
-- Module Name:    VGA - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

use WORK.Led_pckg.ALL;
use WORK.VGA_Fontrom_pckg.ALL;
use WORK.VGASignal_pckg.ALL;
use WORK.vga_clockgen_pckg.ALL;
use WORK.VGA_Fontgen_pckg.ALL;
use WORK.DualPortSram_pckg.ALL;
use WORK.ps2_pckg.ALL;
use WORK.ps2kbdTranslate_pckg.ALL;
use WORK.VGATermEmu_pckg.ALL;
use WORK.VideoRamInterface_pckg.ALL;
use WORK.ECB_BusIntf_pckg.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity VGA is
    generic(
      FREQ_SRAM      : natural := 50_000;	-- operating frequency of dual port ram in KHz
      SRAM_SPEED     : natural := 70;		-- ns, can not be less that 20ns(currently not used)
      SRAM_ADDRWIDTH : natural := 15;		-- SRam Address bits
      SRAM_DATAWIDTH : natural := 16;		-- SRam data bits
      VIDEO_NUMCHARS : natural := 80;
      VIDEO_NUMLINES : natural := 30
    );

    Port ( clk		: in  STD_LOGIC;
           red_out	: out  STD_LOGIC_VECTOR (2 downto 0);
           green_out	: out  STD_LOGIC_VECTOR (2 downto 0);
           blue_out	: out  STD_LOGIC_VECTOR (2 downto 0);
           vsync_out	: out  STD_LOGIC;
           hsync_out	: out  STD_LOGIC;
           reset	: in  STD_LOGIC;
           -- Interface to SDRAM
--           sclkfb : in    std_logic;           -- clock from SDRAM after PCB delays
--           sclk   : out   std_logic;           -- SDRAM clock sync'ed to master clock
--           cke    : out   std_logic;           -- clock-enable to SDRAM
--           cs_n   : out   std_logic;           -- chip-select to SDRAM
--           ras_n  : out   std_logic;           -- SDRAM row address strobe
--           cas_n  : out   std_logic;           -- SDRAM column address strobe
--           we_n   : out   std_logic;           -- SDRAM write enable
--           ba     : out   std_logic_vector(1 downto 0);  -- SDRAM bank address bits
--           sAddr  : out   std_logic_vector(12 downto 0);  -- SDRAM row/column address
--           sData  : inout std_logic_vector(15 downto 0);  -- SDRAM in/out databus
--           dqmh   : out   std_logic;           -- high databits I/O mask
--           dqml   : out   std_logic;           -- low databits I/O mask
           -- Interface to Static RAM
           SROE_n	: out   std_logic;
           SRCE_n	: out   std_logic;
           SRWR_n	: out   std_logic;
           SRData	: inout std_logic_vector(SRAM_DATAWIDTH-1 downto 0);
           SRAddr	: out   std_logic_vector(SRAM_ADDRWIDTH-1 downto 0);
	   -- ps2 keyboard
	   ps2_clk_io  : inout std_logic;	-- PS2 Clock line
           ps2_data_io : inout std_logic;	-- PS2 Data line
           -- diag 7 segment led
           led		: out   std_logic_vector (7 downto 0);
--           p7 : out std_logic;
--           sw3		: in    std_logic            -- switch3
	   ECB_CLK	: in STD_LOGIC;
	   ECB_A	: in STD_LOGIC_VECTOR(7 downto 0);
	   ECB_D	: inout STD_LOGIC_VECTOR(7 downto 0);
--	   ECB_INT	: in STD_LOGIC;
	   ECB_IORQ	: in STD_LOGIC;
--	   ECB_M1	: in STD_LOGIC;
	   ECB_RD	: in STD_LOGIC;
--	   ECB_WAIT	: in STD_LOGIC;
	   ECB_WR	: in STD_LOGIC;
	   testpin1	: out STD_LOGIC;
	   testpin2	: out STD_LOGIC
           );
end VGA;

architecture Behavioral of VGA is
  SIGNAL clk25M,clk50M,clkCharBlink,clkCursorBlink,clkCursorFastBlink : STD_LOGIC;
  SIGNAL ResHighActive : STD_LOGIC;
  SIGNAL GetPixelDataSig : STD_LOGIC;
  SIGNAL CharReady : STD_LOGIC;
  SIGNAL FontRomAddr : STD_LOGIC_VECTOR (11 downto 0);
  SIGNAL FontRomData : STD_LOGIC_VECTOR (7 downto 0);
  SIGNAL CharPixel,ChariRGB : STD_LOGIC_VECTOR (7 downto 0);
  SIGNAL ColCharCount : STD_LOGIC_VECTOR (7 downto 0);
  SIGNAL LinePixCount : STD_LOGIC_VECTOR (9 downto 0);
  
  SIGNAL VideoGenAddr         : STD_LOGIC_VECTOR (SRAM_ADDRWIDTH-1 downto 0);
  SIGNAL VideoGenDataOut      : STD_LOGIC_VECTOR (SRAM_DATAWIDTH-1 downto 0);
  SIGNAL VideoGenStartRead    : STD_LOGIC;
  SIGNAL VideoReadDone        : STD_LOGIC;
  SIGNAL VRDataIn : STD_LOGIC_VECTOR (SRAM_DATAWIDTH-1 downto 0);
  SIGNAL VRDataOut : STD_LOGIC_VECTOR (SRAM_DATAWIDTH-1 downto 0);
  SIGNAL VRStartRead,VRStartWrite,VRDone : STD_LOGIC;
  SIGNAL VRAddr : STD_LOGIC_VECTOR(SRAM_ADDRWIDTH-1 downto 0);
  SIGNAL VRStarted : STD_LOGIC;
  
  SIGNAL VRAccessAllowed : STD_LOGIC;
  
  SIGNAL ps2data_o,ps2data_i : STD_LOGIC_VECTOR (7 downto 0);
  SIGNAL ps2clearInput,ps2setInput,ps2dataAvailable,ps2Busy,ps2Ready : STD_LOGIC;
  SIGNAL ps2CharData_o : STD_LOGIC_VECTOR (7 downto 0);
  SIGNAL ps2CharAvail_o, ps2CharClear_i : STD_LOGIC;
  SIGNAL ps2ParityError,ps2FrameError : STD_LOGIC;
  
  SIGNAL TermEmuReady		: STD_LOGIC;
  SIGNAL CursorVRamAddr		: STD_LOGIC_VECTOR (SRAM_ADDRWIDTH-1 downto 0);
  SIGNAL CursorStyle		: STD_LOGIC_VECTOR (3 downto 0);
  
  SIGNAL HostReceivedChar	: STD_LOGIC_VECTOR (7 downto 0);
  SIGNAL HostCharAvail		: STD_LOGIC;
  SIGNAL LastCharProcessed	: STD_LOGIC_VECTOR (7 downto 0);
  SIGNAL StateHistory		: STD_LOGIC_VECTOR (63 downto 0);
  SIGNAL KeyHistory		: STD_LOGIC_VECTOR (63 downto 0);
  

begin
  --p7 <= GetPixelDataSig;
  --testpin1 <= TermEmuReady; --ps2CharAvail_o;
  --testpin2 <= HostCharAvail; -- ps2CharClear_i;
  
  testpin1 <= ps2CharAvail_o;
  testpin2 <= ps2CharClear_i;
  
  ClockGen: vga_clockgen		-- Clock generator
  PORT MAP(
   reset          => ResHighActive,	-- global reset
   clk            => clk,		-- 100 MHz
   clk50M         => Clk50M,		-- 50 MHz out
   clk25M         => Clk25M,		-- 25 MHz out
   clkCursorBlink => clkCursorBlink,
   clkCursorFastBlink => clkCursorFastBlink,
   clkCharBlink   => clkCharBlink
  );

  -- ps2 interface
  ps2intf: ps2
  PORT MAP(
	clk_i		=> clk25M,
	rst_i		=> Reset,		-- low active
        data_o  	=> ps2data_o,
        data_i  	=> ps2data_i,
        ibf_clr_i	=> ps2clearInput,	-- Ifb flag clear input
        obf_set_i	=> ps2setInput,		-- Obf flag set input
        ibf_o		=> ps2dataAvailable,	-- Received data available
        obf_o		=> open,		-- Data ready to sent

	frame_err_o	=> ps2FrameError,	-- Error receiving data
	parity_err_o	=> ps2ParityError,	-- Error in received data parity
	busy_o		=> ps2Busy,		-- uart busy
	err_clr_i	=> '0',			-- Clear error flags
	wdt_o		=> open,		-- Watchdog timer out every 400uS
	ps2_clk_io	=> ps2_clk_io,		-- physical ps2 port
	ps2_data_io	=> ps2_data_io
  );
  
  ps2Ready <= not ps2Busy;
  ResHighActive <= not reset;
  
  -- ps2 keyboard translation
  kbdtrans: ps2kbdTranslate
    PORT MAP (							-- all Signals active high
	clk		=> clk25M,
	reset		=> resHighActive,
	keyavail_i	=> ps2dataAvailable,			-- in, new scan code available
	keyData_i	=> ps2data_o,				-- in, key scan code
	
	sendData_o	=> ps2data_i,				-- data to be send to keyboard
	sendAvail_o	=> ps2setInput,				-- we have data to be send
	sendFinished_i	=> ps2Ready,				-- from ps2 module, '1' if data has been send
	
	keyClear_o	=> ps2clearInput,			-- out, scancode has been read
	
	CharData_o	=> ps2CharData_o,			-- translated char
	CharAvail_o	=> ps2CharAvail_o,			-- '1' if translated char is available in char_o
	CharClear_i	=> ps2CharClear_i,			-- '1' to indicate CharData_o has been processed
	StateHistory	=> StateHistory,
	KeyHistory	=> KeyHistory);
  
  TermEmu: VGATermEmu
    GENERIC MAP(
      SRAM_ADDRWIDTH => SRAM_ADDRWIDTH,
      SRAM_DATAWIDTH => SRAM_DATAWIDTH,
      VIDEO_NUMCHARS => VIDEO_NUMCHARS,
      VIDEO_NUMLINES => VIDEO_NUMLINES
    )
    PORT MAP 
	 ( clk			=> clk50M,		-- 50 MHz
           reset		=> resHighActive,	-- active high
           VRamAddr		=> VRAddr,
           VRamDataOut		=> VRDataIn,
           VRamDataIn		=> VRDataOut,
           VRamRd		=> VRStartRead, --VRI_rd,
           VRamWr		=> VRStartWrite, --VRI_wr,
           VRamReady		=> VRdone,
	   VRamStarted		=> VRStarted,
           VRamCursorAddress	=> CursorVRamAddr,
           VRamCursorStyle	=> CursorStyle,
           Ready		=> TermEmuReady,
           newCharClk		=> HostCharAvail, --ps2CharAvail_o,
           newChar		=> HostReceivedChar, --ps2CharData_o,
	   newCharProcessed	=> open,  --ps2CharClear_i);
	   LastCharProcessed	=> LastCharProcessed);
 
  -- ECB Interface
  ECB:ECB_BusIntf
  Port Map (
	res_n			=> reset,
	clk			=> clk50M,
	-- ECB Bus
	ECB_clk			=> ECB_CLK,
	rd_n			=> ECB_RD,
	wr_n			=> ECB_WR,
	iorq_n			=> ECB_IORQ,
	A			=> ECB_A,
	D			=> ECB_D,
	-- Keyboard
	KbdChar_i		=> ps2CharData_o,
	KbdCharAvail_i		=> ps2CharAvail_o,
	KbdCharProcessed_o	=> ps2CharClear_i,
	-- Terminal Emulation
	TermChar_o		=> HostReceivedChar,
	TermCharAvail_o		=> HostCharAvail,
	TermReady		=> TermEmuReady,
	
	ps2FrameErr		=> ps2FrameError,
	ps2ParityErr		=> ps2ParityError,
	ps2LowCharAvail		=> ps2dataAvailable,
	StateHistory		=> StateHistory, --(Others => '0'),
	LastCharProcessed	=> LastCharProcessed,
	KeyHistory		=> KeyHistory);



  VGASig:VGASignal			-- generates VGA video signal and fetch signals for pixel data
  PORT MAP(
    reset           => ResHighActive,
    Clk25M          => Clk25M,
    GetPixelDataSig => GetPixelDataSig,	-- 8 25M clocks before data is needed
    ColCharCount    => ColCharCount,
    LinePixCount    => LinePixCount,
    CharPixel       => CharPixel,	-- character pixel data
    ChariRGB        => ChariRGB,	-- RGB,Intensity foreground(4 bit) + RGB,Intensity background(4 bit)
    CharReady       => CharReady,	-- if 1, on rising edge of clk25M, CharPixel and IRGBPixel will be transferred to internal variables
    
    hsync_out       => hsync_out,	-- Signals to VGA Monitor
    vsync_out       => vsync_out,
    red_out         => red_out,
    green_out       => green_out,
    blue_out        => blue_out,
    EnableRamAccess => VRAccessAllowed	-- 1 if access to video ram allowed
  );
  
 
  VGAFont:VGA_Fontrom			-- 8x16 VGA Font, used by VGA_Fontgen
  PORT MAP(
    clk  => clk,
    ena  => '1',
    addr => FontRomAddr,
    dout => FontRomData
  );
  
  VGAFontGenerator: VGA_Fontgen          -- generates pixel data for VGASig
  GENERIC MAP(
    SRAM_ADDRWIDTH => SRAM_ADDRWIDTH,
    SRAM_DATAWIDTH => SRAM_DATAWIDTH
    )
  PORT MAP(
    reset		=> ResHighActive,
    ColCharCount	=> ColCharCount,
    LinePixCount	=> LinePixCount,
    clk			=> clk50M,
    clkCursorBlink	=> clkCursorBlink,
    clkCursorFastBlink	=> clkCursorFastBlink,
    clkCharBlink	=> clkCharBlink,
    GetPixelDataSig	=> GetPixelDataSig,
    CharPixel		=> CharPixel,
    ChariRGB		=> ChariRGB,
    CharReady		=> CharReady,
    CursorVRamAddr	=> CursorVRamAddr,
    CursorStyle		=> CursorStyle,
    VideoRamAddr	=> VideoGenAddr,
    VideoRamData	=> VideoGenDataOut,
    VideoRamEnable	=> VideoGenStartRead,
    VideoReadDone	=> VideoReadDone,
    FontRomAddr		=> FontRomAddr,
    FontRomData		=> FontRomData
    );
    
  VideoRam: DualPortSram		-- dual port video ram, port1 R/O for VGAFontGenerator
  GENERIC MAP(
    FREQ		=> FREQ_SRAM,
    SRAM_SPEED		=> SRAM_SPEED,
    SRAM_ADDRWIDTH	=> SRAM_ADDRWIDTH,
    SRAM_DATAWIDTH	=> SRAM_DATAWIDTH
    )
  PORT MAP(
    clk			=> clk50M,	-- does not work with 100MHz
    reset		=> ResHighActive,
    -- Interface to video character generator
    Addr1		=> VideoGenAddr,
    DataOut1		=> VideoGenDataOut,
    rd1			=> VideoGenStartRead,
    done1		=> VideoReadDone,
    -- Interface to terminal emulation
    DataIn2		=> VRDataIn,
    rd2			=> VRStartRead,
    wr2			=> VRStartWrite,
    Addr2		=> VRAddr,
    DataOut2		=> VRDataOut,
    Done2		=> VRDone,
    Started2		=> VRStarted,
    Enable2		=> VRAccessAllowed,
    -- Interface to Static Ram
    SRAddr		=> SRAddr,
    SRData		=> SRData,
    SR_OE		=> SROE_n,
    SR_CS		=> SRCE_n,
    SR_WR		=> SRWR_n
  );
  
  -- show char on 7 seg display
  led7: Led7seg8
    Port map ( 
	clk => clk50M,
	reset => ResHighActive,
	Din => HostReceivedChar, -- ps2CharData_o,
        DOut => Led);
	
    

END Behavioral;

