----------------------------------------------------------------------------------
-- Engineer:		Armin Diehl
-- Create Date:		00:01:58 01/07/2007 
-- Module Name:		VGATermEmu - Behavioral 
-- Project Name:	VGA Terminal
-- Target Devices:	Spartan 2
-- Tool versions:	ise 8.2.03i
-- Description:		Terminal Emulator for VGA project
--			Emulates a ??? Terminal
-- Revision 0.01 - File Created
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

PACKAGE VGATermEmu_pckg IS
  COMPONENT VGATermEmu is
    generic(
      SRAM_ADDRWIDTH : natural := 15;			-- Ram Address bits
      SRAM_DATAWIDTH : natural := 8;			-- can not be changed without changing the code
      VIDEO_NUMCHARS : natural := 80;
      VIDEO_NUMLINES : natural := 30
    );
    Port ( clk			: in  STD_LOGIC;	-- 50 MHz
           reset		: in  STD_LOGIC;	-- active high
           VRamAddr		: out STD_LOGIC_VECTOR (SRAM_ADDRWIDTH-1 downto 0);
           VRamDataOut		: out STD_LOGIC_VECTOR (15 downto 0);
           VRamDataIn		: in  STD_LOGIC_VECTOR (15 downto 0);
           VRamRd		: out STD_LOGIC;
           VRamWr		: out STD_LOGIC;
           VRamReady		: in  STD_LOGIC;
	   VRamStarted		: in  STD_LOGIC;
           VRamCursorAddress	: out STD_LOGIC_VECTOR (SRAM_ADDRWIDTH-1 downto 0);
           VRamCursorStyle	: out STD_LOGIC_VECTOR (3 downto 0);
           Ready		: out STD_LOGIC;
           newCharClk		: in  STD_LOGIC;
           newChar		: in  STD_LOGIC_VECTOR (7 downto 0);
	   newCharProcessed	: out STD_LOGIC;
	   LastCharProcessed	: out STD_LOGIC_VECTOR (7 downto 0));
    END COMPONENT;
END PACKAGE; 

----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use IEEE.numeric_std.all;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

entity VGATermEmu is
    generic(
      SRAM_ADDRWIDTH : natural := 15;			-- Ram Address bits
      SRAM_DATAWIDTH : natural := 8;			-- can not be changed without changing the code
      VIDEO_NUMCHARS : natural := 80;
      VIDEO_NUMLINES : natural := 30
    );
    Port ( clk			: in  STD_LOGIC;	-- 50 MHz
           reset		: in  STD_LOGIC;	-- active high
           VRamAddr		: out STD_LOGIC_VECTOR (SRAM_ADDRWIDTH-1 downto 0);
           VRamDataOut		: out STD_LOGIC_VECTOR (15 downto 0);
           VRamDataIn		: in  STD_LOGIC_VECTOR (15 downto 0);
           VRamRd		: out STD_LOGIC;
           VRamWr		: out STD_LOGIC;
           VRamReady		: in  STD_LOGIC;
	   VRamStarted		: in  STD_LOGIC;
           VRamCursorAddress	: out STD_LOGIC_VECTOR (SRAM_ADDRWIDTH-1 downto 0);
           VRamCursorStyle	: out STD_LOGIC_VECTOR (3 downto 0);
           Ready		: out STD_LOGIC;
           newCharClk		: in  STD_LOGIC;
           newChar		: in  STD_LOGIC_VECTOR (7 downto 0);
	   newCharProcessed	: out STD_LOGIC;
	   LastCharProcessed	: out STD_LOGIC_VECTOR (7 downto 0));
end VGATermEmu;

architecture Behavioral of VGATermEmu is
TYPE TState IS (S_ClrScr,
		S_Idle,
		S_AnalyzeChar,
		S_WriteChar,		-- write to curr addr and inc addr
		S_WriteChar1,S_WriteChar2,
		S_WaitCharGone,
		S_WaitCharGone1,
		S_WaitCharGone2,
		S_memFill,S_memFill_1,S_memFill_2,
		S_memMoveUp,S_memMoveUp_1,S_memMoveUp_2,S_memMoveUp_3,S_memMoveUp_4,
		S_memMoveDn,S_memMoveDn_1,S_memMoveDn_2,S_memMoveDn_3,S_memMoveDn_4,
		S_CursorRight,S_CursorRight_1,S_CursorLeft,S_CursorUp,
		S_CursorDown,S_CursorDown_1,
		S_CursorX0,S_gotoX,S_gotoY,S_gotoY_1,
		S_ClrLine,S_ClrEol,S_ClrEos,
		S_InsChar,S_InsChar_1,S_InsLine,
		S_DelChar,S_DelChar_1,S_DelLine,S_ClearLastLine,
		S_ClrFore,S_ClrFore_1,S_ClrFore_2,S_ClrFore_3,S_ClrFore_4,S_ClrFore_5);
SIGNAL State		: TState;
constant HI		: std_logic := '1';
constant LO		: std_logic := '0';
constant SCREENCHARS    : natural := VIDEO_NUMCHARS * VIDEO_NUMLINES;
--constant SCREENBYTES	: natural := SCREENCHARS * 2;
constant LASTLINESTARTADDR  : natural := VIDEO_NUMCHARS * (VIDEO_NUMLINES-1);
CONSTANT C_CG		: STD_LOGIC_VECTOR(7 downto 0) := x"07";	-- BEL
CONSTANT C_CH		: STD_LOGIC_VECTOR(7 downto 0) := x"08";	-- ^H
CONSTANT C_CI		: STD_LOGIC_VECTOR(7 downto 0) := x"09";	-- ^I
CONSTANT C_CJ		: STD_LOGIC_VECTOR(7 downto 0) := x"0A";
CONSTANT C_CK		: STD_LOGIC_VECTOR(7 downto 0) := x"0B";
CONSTANT C_CL		: STD_LOGIC_VECTOR(7 downto 0) := x"0C";
CONSTANT C_CM		: STD_LOGIC_VECTOR(7 downto 0) := x"0D";
CONSTANT C_CN		: STD_LOGIC_VECTOR(7 downto 0) := x"0E";
CONSTANT C_CO		: STD_LOGIC_VECTOR(7 downto 0) := x"0F";
CONSTANT C_CP		: STD_LOGIC_VECTOR(7 downto 0) := x"10";
CONSTANT C_CQ		: STD_LOGIC_VECTOR(7 downto 0) := x"11";
CONSTANT C_CR		: STD_LOGIC_VECTOR(7 downto 0) := x"12";
CONSTANT C_CS		: STD_LOGIC_VECTOR(7 downto 0) := x"13";
CONSTANT C_CT		: STD_LOGIC_VECTOR(7 downto 0) := x"14";
CONSTANT C_CU		: STD_LOGIC_VECTOR(7 downto 0) := x"15";
CONSTANT C_CV		: STD_LOGIC_VECTOR(7 downto 0) := x"16";
CONSTANT C_CW		: STD_LOGIC_VECTOR(7 downto 0) := x"17";
CONSTANT C_CX		: STD_LOGIC_VECTOR(7 downto 0) := x"18";
CONSTANT C_CY		: STD_LOGIC_VECTOR(7 downto 0) := x"19";
CONSTANT C_CZ		: STD_LOGIC_VECTOR(7 downto 0) := x"1A";
CONSTANT C_ESC		: STD_LOGIC_VECTOR(7 downto 0) := x"1B";
CONSTANT C_SPACE	: STD_LOGIC_VECTOR(7 downto 0) := x"20";
CONSTANT C_STAR		: STD_LOGIC_VECTOR(7 downto 0) := x"2A";	-- *
CONSTANT C_PLUS		: STD_LOGIC_VECTOR(7 downto 0) := x"2B";	-- +
CONSTANT C_SEMICOLON	: STD_LOGIC_VECTOR(7 downto 0) := x"3B";	-- ;
CONSTANT C_COLON	: STD_LOGIC_VECTOR(7 downto 0) := x"3A";	-- :
CONSTANT C_POINT	: STD_LOGIC_VECTOR(7 downto 0) := x"2E";	-- .
CONSTANT C_EQUAL	: STD_LOGIC_VECTOR(7 downto 0) := x"3D";	-- =

CONSTANT C_C		: STD_LOGIC_VECTOR(7 downto 0) := x"43";	-- C
CONSTANT C_smallC	: STD_LOGIC_VECTOR(7 downto 0) := x"63";	-- c

CONSTANT C_E		: STD_LOGIC_VECTOR(7 downto 0) := x"45";	-- E
CONSTANT C_F		: STD_LOGIC_VECTOR(7 downto 0) := x"46";	-- F
CONSTANT C_G		: STD_LOGIC_VECTOR(7 downto 0) := x"47";	-- G

CONSTANT C_P		: STD_LOGIC_VECTOR(7 downto 0) := x"50";	-- P
CONSTANT C_Q		: STD_LOGIC_VECTOR(7 downto 0) := x"51";	-- Q
CONSTANT C_R		: STD_LOGIC_VECTOR(7 downto 0) := x"52";	-- R
CONSTANT C_S		: STD_LOGIC_VECTOR(7 downto 0) := x"53";	-- S
CONSTANT C_T		: STD_LOGIC_VECTOR(7 downto 0) := x"54";	-- T
CONSTANT C_smallT	: STD_LOGIC_VECTOR(7 downto 0) := x"74";	-- t
CONSTANT C_U		: STD_LOGIC_VECTOR(7 downto 0) := x"55";
CONSTANT C_V		: STD_LOGIC_VECTOR(7 downto 0) := x"56";
CONSTANT C_W		: STD_LOGIC_VECTOR(7 downto 0) := x"57";
CONSTANT C_X		: STD_LOGIC_VECTOR(7 downto 0) := x"58";
CONSTANT C_Y		: STD_LOGIC_VECTOR(7 downto 0) := x"59";	-- Y
CONSTANT C_smallY	: STD_LOGIC_VECTOR(7 downto 0) := x"79";	-- y
CONSTANT C_Z		: STD_LOGIC_VECTOR(7 downto 0) := x"5A";	-- Z
CONSTANT C_openBra	: STD_LOGIC_VECTOR(7 downto 0) := x"28";	-- (
CONSTANT C_closeBra	: STD_LOGIC_VECTOR(7 downto 0) := x"29";	-- )
CONSTANT C_0		: STD_LOGIC_VECTOR(7 downto 0) := x"30";	-- 0
CONSTANT C_1		: STD_LOGIC_VECTOR(7 downto 0) := x"31";
CONSTANT C_2		: STD_LOGIC_VECTOR(7 downto 0) := x"32";
CONSTANT C_3		: STD_LOGIC_VECTOR(7 downto 0) := x"33";
CONSTANT C_4		: STD_LOGIC_VECTOR(7 downto 0) := x"34";
CONSTANT C_5		: STD_LOGIC_VECTOR(7 downto 0) := x"35";
CONSTANT C_6		: STD_LOGIC_VECTOR(7 downto 0) := x"36";
CONSTANT C_7		: STD_LOGIC_VECTOR(7 downto 0) := x"37";
CONSTANT C_8		: STD_LOGIC_VECTOR(7 downto 0) := x"38";
CONSTANT C_9		: STD_LOGIC_VECTOR(7 downto 0) := x"39";



CONSTANT CursorBlinkSlowBit : NATURAL := 0;
CONSTANT CursorBlinkFastBit : NATURAL := 1;
CONSTANT CursorBlockBit     : NATURAL := 2;
CONSTANT CursorLineBit      : NATURAL := 3;
SIGNAL VRamCursAddr	: STD_LOGIC_VECTOR(SRAM_ADDRWIDTH-1 downto 0);
SIGNAL currChar,currAttr: STD_LOGIC_VECTOR(7 downto 0);
SIGNAL escActive	: STD_LOGIC;
SIGNAL escParamCount	: INTEGER RANGE 0 TO 2;
SIGNAL escChar		: STD_LOGIC_VECTOR(7 downto 0);
-- for memory operations (fill,move)
SIGNAL memCount		: INTEGER RANGE 0 TO SCREENCHARS+1;
SIGNAL memAddr,memAddr2	: STD_LOGIC_VECTOR(SRAM_ADDRWIDTH-1 downto 0);
SIGNAL memNextState,returnState	: TState;	-- state after memory operation

begin
  VRamCursorAddress <= VRamCursAddr;
  
  ProcessChar:PROCESS (Reset,clk)
  VARIABLE currX,newX : INTEGER RANGE 0 TO VIDEO_NUMCHARS-1;
  VARIABLE currY,newY : INTEGER RANGE 0 TO VIDEO_NUMLINES-1;
  BEGIN
    IF Reset = HI THEN
      State <= S_ClrScr;		-- at reset, clear the screen
      currAttr <= "00001111";
      Ready <= HI;
      VRamCursorStyle <= "0101";	-- slow blink, block
      VRamRd <= LO;
      VRamWr <= LO;
      VRamCursAddr <= (Others => LO); currX := 0; currY := 0;
      escActive <= LO;
      escParamCount <= 0;
    ELSIF Rising_Edge(Clk) THEN
      CASE State IS
        WHEN S_Idle		=> newCharProcessed <= LO;
				   Ready <= HI;
				   returnState <= S_WaitCharGone;
				   IF newCharClk = HI THEN
				     currChar <= newChar;
				     LastCharProcessed <= newChar;
				     State <= S_AnalyzeChar;
				     --Ready <= LO;
				     IF (EscParamCount = 0) THEN
				       escChar <= newChar;
				     END IF;
				   END IF;
	WHEN S_AnalyzeChar	=> Ready <= LO;
				   IF escActive = HI THEN
				     CASE escChar IS
				       WHEN C_STAR	=> State <= S_ClrScr; escActive <= LO;
				       WHEN C_COLON	=> State <= S_ClrScr; escActive <= LO;
				       WHEN C_PLUS	=> State <= S_ClrFore; escActive <= LO;
				       WHEN C_SEMICOLON	=> State <= S_ClrFore; escActive <= LO;
				       WHEN C_E		=> State <= S_InsLine; escActive <= LO;
				       WHEN C_R		=> State <= S_DelLine; escActive <= LO;
				       WHEN C_Q		=> State <= S_InsChar; escActive <= LO;
				       WHEN C_T		=> State <= S_ClrEol; escActive <= LO;
				       WHEN C_W		=> State <= S_DelChar; escActive <= LO;
				       WHEN C_smallT	=> State <= S_ClrEol; escActive <= LO;
				       WHEN C_Y		=> State <= S_ClrEos; escActive <= LO;
				       WHEN C_smallY	=> State <= S_ClrEos; escActive <= LO;
				       				       
				       WHEN C_openBra	=> currAttr(3) <= '1';
							   State <= returnState; escActive <= LO;
				       WHEN C_closeBra	=> currAttr(3) <= '0';
							   State <= returnState; escActive <= LO;
				       WHEN C_G		=> IF escParamCount = 1 THEN
							     escActive <= LO;
							     escParamCount <= 0;
							     State <= returnState;
							     IF currChar(0) = '1' THEN		-- bit0 -> Invisible, bit1 -> Blink
							       currAttr <= currAttr(7) & "000" & currAttr(3) & "000";		-- black on black, keep intensity bit
							     ELSE
							       -- Bit4 -> underline, ignore for now
							       -- Bit2 -> reverse
							       IF currChar(2) = '1' THEN
							         currAttr <= currAttr(7) & "111" & currAttr(3) & "000";
							       ELSE
							         currAttr <= currAttr(7) & "000" & currAttr(3) & "111";
							       END IF;
							     END IF;
							   ELSE
							     escParamCount <= escParamCount + 1;
							     State <= returnState;
							   END IF;
				       WHEN C_C		=> IF escParamCount = 1 THEN  -- foreground color
							     currAttr <= currAttr(7 downto 3) & currChar(2 downto 0);
							     escActive <= LO;
							     escParamCount <= 0;
							     State <= returnState;
							   ELSE
							     escParamCount <= escParamCount + 1;
							     State <= returnState;
							   END IF;
				       WHEN C_smallC	=> IF escParamCount = 1 THEN  -- background color
							     currAttr <= currAttr(7) & currChar(2 downto 0) & currAttr(3 downto 0);
							     escActive <= LO;
							     escParamCount <= 0;
							     State <= returnState;
							   ELSE
							     escParamCount <= escParamCount + 1;
							     State <= returnState;
							   END IF;
				       WHEN C_POINT	=> IF escParamCount = 1 THEN
							     escActive <= LO;
							     -- "0":no cursor, "1":Blink, 2:non Blink block, 
							     -- 3:Blinking line, 4: steady line
							     -- 5: same as 1 (wyse)
							     -- 6: fast blinking block, 7:fast blinking line
							     CASE currChar IS
							       WHEN x"30" => VRamCursorStyle <= "0000";
							       WHEN x"31" => VRamCursorStyle <= "0101";
							       WHEN x"32" => VRamCursorStyle <= "0100";
							       WHEN x"33" => VRamCursorStyle <= "1001";
							       WHEN x"34" => VRamCursorStyle <= "1000";
							       WHEN x"35" => VRamCursorStyle <= "0101";
							       WHEN x"36" => VRamCursorStyle <= "0110";
							       WHEN x"37" => VRamCursorStyle <= "1010";
							       WHEN Others=> NULL;
							     END CASE;
							     escParamCount <= 0;
							     State <= returnState;
							   ELSE
							     escParamCount <= escParamCount + 1;
							     State <= returnState;
							   END IF;
				       WHEN C_EQUAL	=> IF escParamCount = 1 THEN	-- Y
							     escParamCount <= escParamCount + 1;
							     newY := conv_integer(currChar - x"20");
							     State <= S_gotoY;
							   ELSIF escParamCount = 2 THEN	-- X
							     escParamCount <= 0;
							     escActive <= LO;
							     newX := conv_integer(currChar - x"20");
							     State <= S_gotoX;
							   ELSE
							     escParamCount <= escParamCount + 1;
							     State <= returnState;
							   END IF;
				       WHEN others	=> escActive <= LO;
				     END CASE;
				   ELSE
				     CASE currChar IS		-- hope this works with a variable
				       WHEN C_ESC	=> escActive <= HI;
							   escParamCount <= 0;
							   State <= S_WaitCharGone;
				       WHEN C_CH	=> State <= S_CursorLeft;
				       WHEN C_CG	=> State <= returnState;	-- ignore BEL for now (no speaker yet)
				       WHEN C_CK	=> State <= S_CursorUp;
				       WHEN C_CV	=> State <= S_CursorDown;
				       WHEN C_CJ	=> State <= S_CursorDown;
				       WHEN C_CM	=> State <= S_CursorX0;
				       WHEN x"0F"	=> -- cursor to col 0 and down
							   VRamCursAddr <= VRamCursAddr - currX;
							   currX := 0;
							   State <= S_CursorDown;
				       WHEN x"1E"	=> VRamCursAddr <= (Others => '0');
							   currX := 0; currY := 0;
							   State <= returnState;
				       WHEN C_CZ	=> State <= S_ClrScr;
				       WHEN C_CL	=> State <= S_ClrScr;
				       WHEN Others	=> State <= S_WriteChar;
				     END CASE;
				   END IF;
	WHEN S_WriteChar	=> VRamAddr <= VRamCursAddr;	-- write to video ram
				   VRamDataOut(15 downto 8) <= currChar;
				   VRamDataOut(7 downto 0) <= currAttr;
				   VRamWr <= HI;
				   State <= S_WriteChar1;
	WHEN S_WriteChar1	=> IF VRamStarted = '1' THEN
				     State <= S_WriteChar2;
				   END IF;
	WHEN S_WriteChar2	=> IF VRamReady = HI THEN
				     VRamWr <= LO;
				     returnState <= S_WaitCharGone;
				     State <= S_CursorRight;	-- cursor right and check scroll
				   END IF;
				   
	WHEN S_WaitCharGone	=> newCharProcessed <= HI;	-- set processed flag
				   State <= S_WaitCharGone1;
	WHEN S_WaitCharGone1	=> State <= S_WaitCharGone2;	-- for at least 2 clocks
	WHEN S_WaitCharGone2	=> IF newCharClk = LO THEN	-- if the new char signal has gone
--				     Ready <= HI;		-- we are ready for the next char
				     State <= S_Idle;
				   END IF;
	WHEN S_ClrScr		=> memNextState <= S_WaitCharGone;
				   memAddr <= (Others => '0');	-- start addr 0
				   memCount <= SCREENCHARS;	-- num of chars
				   currAttr <= "00001111";
				   currChar <= C_SPACE;
				   currX := 0; currY := 0; VRamCursAddr <= (Others => '0');
				   State <= S_memFill;
				   
	-- fill memory at memCount@memAddr with currChar and currAttr
	-- memcount is number of characters(16 Bit)
	WHEN S_memFill		=> IF memCount = 0 THEN
				     State <= memNextState;
				   ELSE
				     VRamAddr <= memAddr;
				     VRamDataOut(15 downto 8) <= currChar;
				     VRamDataOut(7 downto 0) <= currAttr;
				     VRamWr <= HI;
				     State <= S_memFill_1;
				   END IF;
	WHEN S_memFill_1	=> IF VRamStarted = HI THEN
				     State <= S_memFill_2;
				   END IF;
	WHEN S_memFill_2	=> IF VRamReady = HI THEN
				     VRamWr <= LO;
				     memCount <= memCount - 1;
				     State <= S_memFill;
				     memAddr <= memAddr + 1;
				   END IF;
				   
	-- move count memory from memAddr to memAddr2, addr will be incremented
	-- memcount is number of characters(16 Bit)
	WHEN S_memMoveUp	=> IF memCount = 0 THEN
				     State <= memNextState;
				   ELSE
				     VRamAddr <= memAddr;
				     VRamRd <= HI;
				     State <= S_memMoveUp_1;
				   END IF;
	WHEN S_memMoveUp_1	=> IF VRamReady = HI THEN
				     VRamRd <= LO;
				     VRamDataOut <= VRamDataIn;
				     memAddr <= memAddr + 1;
				     VRamAddr <= memAddr2;
				     State <= S_memMoveUp_2;
				   END IF;
	WHEN S_memMoveUp_2	=> IF VRamReady = LO THEN
				     State <= S_memMoveUp_3;
				     VRamWr <= HI;
				   END IF;
	WHEN S_memMoveUp_3	=> IF VRamReady = HI THEN
				     VRamWr <= LO;
				     memAddr2 <= memAddr2 + 1;
				     memCount <= memCount - 1;
				     State <= S_memMoveUp_4;
				   END IF;
	WHEN S_memMoveUp_4	=> IF VRamReady = LO THEN
				     State <= S_memMoveUp;
				   END IF;
				   
	-- clear foreground characters
	WHEN S_ClrFore		=> memCount <= SCREENCHARS;
				   memAddr <= (Others => '0');
				   State <= S_ClrFore_1;
	WHEN S_ClrFore_1	=> IF memCount = 0 THEN
				     State <= returnState;
				   ELSE
				     VRamAddr <= memAddr;
				     VRamRd <= HI;
				     State <= S_ClrFore_2;
				   END IF;
	WHEN S_ClrFore_2	=> IF VRamReady = HI THEN
				     VRamRd <= LO;
				     IF VRamDataIn(3) = '1' THEN	-- intensity bit
				       VRamDataOut <= x"20" & "00001111";
				       State <= S_ClrFore_3;
				     ELSE
				       State <= S_ClrFore_5;
				     END IF;
				   END IF;
	WHEN S_ClrFore_3	=> IF VRamReady = LO THEN
				     State <= S_ClrFore_4;
				     VRamWr <= HI;
				   END IF;
	WHEN S_ClrFore_4	=> IF VRamReady = HI THEN
				     VRamWr <= LO;
				     State <= S_ClrFore_5;
				   END IF;
	WHEN S_ClrFore_5	=> IF VRamReady = LO THEN
				     memAddr <= memAddr + 1;
				     memCount <= memCount - 1;
				     State <= S_ClrFore_1;
				   END IF;
				   
	-- move count memory from memAddr to memAddr2, addr will be decremented
	-- memcount is number of characters(16 Bit)
	WHEN S_memMoveDn	=> IF memCount = 0 THEN
				     State <= memNextState;
				   ELSE
				     VRamAddr <= memAddr;
				     VRamRd <= HI;
				     State <= S_memMoveDn_1;
				   END IF;
	WHEN S_memMoveDn_1	=> IF VRamReady = HI THEN
				     VRamRd <= LO;
				     VRamDataOut <= VRamDataIn;
				     memAddr <= memAddr - 1;
				     VRamAddr <= memAddr2;
				     State <= S_memMoveDn_2;
				   END IF;
	WHEN S_memMoveDn_2	=> IF VRamReady = LO THEN
				     State <= S_memMoveDn_3;
				     VRamWr <= HI;
				   END IF;
	WHEN S_memMoveDn_3	=> IF VRamReady = HI THEN
				     VRamWr <= LO;
				     memAddr2 <= memAddr2 - 1;
				     memCount <= memCount - 1;
				     State <= S_memMoveDn_4;
				   END IF;
	WHEN S_memMoveDn_4	=> IF VRamReady = LO THEN
				     State <= S_memMoveDn;
				   END IF;
			
	-- moves cursor right (with scroll), next state is returnState
	WHEN S_CursorRight	=> IF currX = VIDEO_NUMCHARS-1 THEN
				     IF CurrY = VIDEO_NUMLINES-1 THEN
				       memAddr2 <= (Others => '0');	-- scroll
				       memAddr <= std_logic_vector(to_unsigned(VIDEO_NUMCHARS,memAddr'LENGTH));
				       memCount <= SCREENCHARS-VIDEO_NUMCHARS;
				       memNextState <= S_CursorRight_1;
				       State <= S_memMoveUp;
				     ELSE	-- next line
				       currX := 0;
				       currY := currY + 1;
				       VRamCursAddr <= VRamCursAddr + 1;
				       State <= returnState;
				     END IF;
				   ELSE		-- right in column
				     currX := currX + 1;
				     VRamCursAddr <= VRamCursAddr + 1;
				     State <= returnState;
				   END IF;
	
	WHEN S_CursorRight_1	=> VRamCursAddr <= std_logic_vector(to_unsigned(LASTLINESTARTADDR,VRamCursAddr'LENGTH));
				   currX := 0;
				   currY := VIDEO_NUMLINES-1;
				   State <= S_ClrLine;			-- clear last line after scroll
				   
	WHEN S_CursorDown	=> IF CurrY = VIDEO_NUMLINES-1 THEN	-- move cursor down with scroll
				     memAddr2 <= (Others => '0');	-- scroll
				     memAddr <= std_logic_vector(to_unsigned(VIDEO_NUMCHARS,memAddr2'LENGTH));
				     memCount <= SCREENCHARS-VIDEO_NUMCHARS;
				     memNextState <= S_CursorDown_1;
				     State <= S_memMoveUp;
				   ELSE					-- no scroll, inc cursorPos by one line
				     VRamCursAddr <= VRamCursAddr + VIDEO_NUMCHARS;
				     currY := currY + 1;
				     State <= returnState;
				   END IF;
				   
	WHEN S_CursorDown_1	=> -- after scroll
				   State <= S_ClrLine;
				   
	WHEN S_CursorLeft	=> IF (currY /= 0) OR (currX /= 0) THEN
				     VRamCursAddr <= VRamCursAddr-1;
				     IF currX = 0 THEN
				       currX := VIDEO_NUMCHARS-1;
				       currY := currY - 1;
				     ELSE
				       currX := currX - 1;
				     END IF;
				   END IF;
				   State <= returnState;
	WHEN S_CursorUp		=> IF currY > 0 THEN
				     currY := currY - 1;
				     VRamCursAddr <= VRamCursAddr - VIDEO_NUMCHARS;
				   END IF;
				   State <= returnState;
	WHEN S_CursorX0		=> VRamCursAddr <= VRamCursAddr - currX;
				   currX := 0;
				   State <= returnState;
	WHEN S_gotoX		=> IF newX > VIDEO_NUMCHARS-1 THEN
				     VRamCursAddr <= VRamCursAddr - currX + VIDEO_NUMCHARS;
				     currX := VIDEO_NUMCHARS-1;
				   ELSE
				     VRamCursAddr <= VRamCursAddr - currX + newX;
				     currX := newX;
				   END IF;
				   State <= returnState;
	WHEN S_gotoY		=> VRamCursAddr <= conv_std_logic_vector (currX,VRamCursAddr'LENGTH);	-- first line same col
				   currY := 0;
				   State <= S_gotoY_1;
	WHEN S_gotoY_1		=> IF newY = 0 THEN
				     State <= returnState;
				   ELSE
				     newY := newY - 1;
				     VRamCursAddr <= VRamCursAddr + VIDEO_NUMCHARS;
				     currY := currY + 1;
				   END IF;
	WHEN S_ClrLine		=> -- clear current line from beginning to end
				   memAddr <= VRamCursAddr - currX;
				   memCount <= VIDEO_NUMCHARS;
				   currChar <= x"20";
				   memNextState <= returnState;
				   State <= S_MemFill;
	WHEN S_ClrEol		=> -- clear to end of line
				   memCount <= VIDEO_NUMCHARS - currX;
				   currChar <= x"20";
				   memAddr <= VRamCursAddr;
				   memNextState <= returnState;
				   State <= S_MemFill;
	WHEN S_ClrEos		=> -- clear to end of screen
				   memCount <= SCREENCHARS - conv_integer(VRamCursAddr);
				   currChar <= x"20";
				   memAddr <= VRamCursAddr;
				   memNextState <= returnState;
				   State <= S_MemFill;
	WHEN S_InsChar		=> memCount <=  VIDEO_NUMCHARS - currX - 1;
				   memAddr <= VRamCursAddr - currX + VIDEO_NUMCHARS - 2;
				   memAddr2 <= VRamCursAddr - currX + VIDEO_NUMCHARS - 1;
				   memNextState <= S_InsChar_1;
				   State <= S_MemMoveDn;
	WHEN S_InsChar_1	=> memCount <= 1;		-- write one blank @ cursor pos
				   currChar <= x"20";
				   memAddr <= VRamCursAddr;
				   memNextState <= returnState;
				   State <= S_MemFill;
	WHEN S_DelChar		=> memCount <=  VIDEO_NUMCHARS - currX - 1;
				   memAddr <= VRamCursAddr + 1;
				   memAddr2 <= VRamCursAddr;
				   memNextState <= S_DelChar_1;
				   State <= S_MemMoveUp;
	WHEN S_DelChar_1	=> memCount <= 1;		-- write one blank @ last char of line
				   currChar <= x"20";
				   memAddr <= VRamCursAddr-currX+VIDEO_NUMCHARS - 1;
				   memNextState <= returnState;
				   State <= S_MemFill;
	WHEN S_InsLine		=> IF CurrY = VIDEO_NUMLINES-1 THEN			-- if in last line
				     VRamCursAddr <= VRamCursAddr - currX;		-- goto col0
				     currX := 0;
				     State <= S_ClrLine;
				   ELSE
				     VRamCursAddr <= VRamCursAddr - currX;		-- goto col0
				     currX := 0;
				     --memCount <= (SCREENCHARS - conv_integer(VRamCursAddr) - currX - currX) / 2;
				     memCount <= SCREENCHARS - conv_integer(VRamCursAddr) - currX;
				     memAddr <=conv_std_logic_vector (SCREENCHARS-VIDEO_NUMCHARS,memAddr'LENGTH);
				     memAddr2 <= conv_std_logic_vector (SCREENCHARS,memAddr2'LENGTH);
				     memNextState <= S_ClrLine;
				     State <= S_MemMoveDn;
				   END IF;
	WHEN S_DelLine		=> VRamCursAddr <= VRamCursAddr - currX;		-- goto col0
				   currX := 0;
				   IF CurrY = VIDEO_NUMLINES-1 THEN			-- if in last line
				     State <= S_ClrLine;
				   ELSE
				     memAddr2 <= VRamCursAddr - currX;
				     memAddr <= VRamCursAddr - currX + VIDEO_NUMCHARS;
				     --memCount <= SCREENCHARS - (conv_integer(VRamCursAddr)/2) - currX;
				     memCount <= SCREENCHARS - conv_integer(VRamCursAddr) - currX;
				     memNextState <= S_ClearLastLine;
				     State <= S_memMoveUp;
				   END IF;
	WHEN S_ClearLastLine	=> memAddr <= conv_std_logic_vector (SCREENCHARS-VIDEO_NUMCHARS,memAddr'LENGTH);
				   memCount <= VIDEO_NUMCHARS;
				   currChar <= x"20";
				   memNextState <= returnState;
				   State <= S_memFill;
      END CASE;
    END IF;
  END PROCESS;
  
end Behavioral;

