----------------------------------------------------------------------------------
-- Engineer:		Armin Diehl
-- Create Date:		14:30:29 01/13/2007 
-- Module Name:		VideoRamInterface - Behavioral 
-- Project Name:	VGA
-- Target Devices:	XC2S200
-- Tool versions:	ISE 8.2.03i
-- Description: 	16 Bit Interface to Video Ram
-- Revision 0.01 - File Created
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;


PACKAGE VideoRamInterface_pckg IS
    COMPONENT VideoRamInterface is
    generic(
	SRAM_ADDRWIDTH : natural := 15;       -- Ram Address bits
	SRAM_DATAWIDTH : natural := 8
    );
    Port ( 
	clk		: in  STD_LOGIC;
	reset		: in  STD_LOGIC;		-- high active
	rd		: in  STD_LOGIC;
	wr		: in  STD_LOGIC;
	done		: out STD_LOGIC;
	DataRd		: out STD_LOGIC_VECTOR(15 downto 0);
	Datawr		: in  STD_LOGIC_VECTOR(15 downto 0);
	Addr		: in  STD_LOGIC_VECTOR(SRAM_ADDRWIDTH-1 downto 0);
	
	-- interface to dual port RAM
	VRrd		: out STD_LOGIC;
	VRwr		: out STD_LOGIC;
	VRstarted	: in STD_LOGIC;
	VRDone		: in STD_LOGIC;
	VRAddr		: out STD_LOGIC_VECTOR(SRAM_ADDRWIDTH-1 downto 0);
	VRDataRead	: in STD_LOGIC_VECTOR(SRAM_DATAWIDTH-1 downto 0);
	VRDataWrite	: out STD_LOGIC_VECTOR(SRAM_DATAWIDTH-1 downto 0)
    );
    END COMPONENT;
END PACKAGE;

----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity VideoRamInterface is
    generic(
	SRAM_ADDRWIDTH : natural := 15;       -- Ram Address bits
	SRAM_DATAWIDTH : natural := 8
    );
    Port ( 
	clk		: in  STD_LOGIC;
	reset		: in  STD_LOGIC;		-- high active
	rd		: in  STD_LOGIC;
	wr		: in  STD_LOGIC;
	done		: out STD_LOGIC;
	DataRd		: out STD_LOGIC_VECTOR(15 downto 0);
	Datawr		: in  STD_LOGIC_VECTOR(15 downto 0);
	Addr		: in  STD_LOGIC_VECTOR(SRAM_ADDRWIDTH-1 downto 0);
	
	-- interface to dual port RAM
	VRrd		: out STD_LOGIC;
	VRwr		: out STD_LOGIC;
	VRstarted	: in STD_LOGIC;
	VRDone		: in STD_LOGIC;
	VRAddr		: out STD_LOGIC_VECTOR(SRAM_ADDRWIDTH-1 downto 0);
	VRDataRead	: in STD_LOGIC_VECTOR(SRAM_DATAWIDTH-1 downto 0);
	VRDataWrite	: out STD_LOGIC_VECTOR(SRAM_DATAWIDTH-1 downto 0)
	
    );
end VideoRamInterface;

architecture Behavioral of VideoRamInterface is
TYPE TState IS (S_Idle,S_Start1,S_WaitStart1,S_WaitDone1,S_Done1,
		S_WaitStart2,S_WaitDone2,S_Done2);
SIGNAL State : TState;
SIGNAL isWrite : STD_LOGIC;
begin

  Main:PROCESS(clk,reset)
  BEGIN
    IF Rising_Edge(Clk) THEN
      IF Reset = '1' THEN
        VRrd <= '0'; VRwr <= '0'; done <= '0';
      END IF;
      CASE State IS
        WHEN S_Idle =>		IF (rd = '1') OR (wr = '1') THEN
				  isWrite <= wr;
				  VRAddr <= Addr;
				  VRDataWrite <= DataWr(15 downto 8);	-- does not matter if set for read
				  State <= S_Start1;
				END IF;
				Done <= '0';
	WHEN S_Start1 =>	VRrd <= not isWrite;
				VRwr <= isWrite;
				State <= S_WaitStart1;
	WHEN S_WaitStart1 =>	IF VRstarted = '1' THEN
				  State <= S_WaitDone1;
				END IF;
	WHEN S_WaitDone1 =>	IF VRDone = '1' THEN
				  State <= S_Done1;
				  VRrd <= '0'; VRwr <= '0';
				  VRAddr <= Addr+1;
				  IF isWrite = '0' THEN
				    DataRd(15 downto 8) <= VRDataRead;
				  END IF;
				END IF;
	WHEN S_Done1 =>		State <= S_WaitStart2;
				VRDataWrite <= DataWr(7 downto 0);	-- does not matter if set for read
				VRrd <= not isWrite;
				VRwr <= isWrite;
	WHEN S_WaitStart2 =>	IF VRStarted = '1' THEN
				  State <= S_WaitDone2;
				END IF;
	WHEN S_WaitDone2 =>	IF VRDone = '1' THEN
				  State <= S_Done2;
				  Done <= '1';
				  VRrd <= '0'; VRwr <= '0';
				  IF isWrite = '0' THEN
				    DataRd(7 downto 0) <= VRDataRead;
				  END IF;
				END IF;
	WHEN S_Done2 => 	State <= S_Idle;
      END CASE;
    END IF;
  END PROCESS;

end Behavioral;

