----------------------------------------------------------------------------------
-- Engineer: 		Armin Diehl
-- Create Date:		10:25:01 02/11/2007 
-- Module Name:		ps2kbdTranslate - Behavioral 
-- Target Devices:	XC2S200
-- Tool versions:	ise 9.1.01i
-- Description:		converts ps2 keyboard scancodes to ascii (German Keyboard)
-- Revision 0.01 - File Created
--
-- TODO: leds not working yet
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;


PACKAGE ps2kbdTranslate_pckg IS
  COMPONENT ps2kbdTranslate is
    PORT (							-- all Signals active high
	clk		: in  STD_LOGIC;
	reset		: in  STD_LOGIC;			-- high active
	keyavail_i	: in  STD_LOGIC;			-- in, new key available
	
	sendData_o	: out STD_LOGIC_VECTOR (7 downto 0);	-- data to be send to keyboard
	sendAvail_o	: out STD_LOGIC;			-- we have data to be send
	sendFinished_i	: in  STD_LOGIC;			-- from ps2 module, '1' if data has been send

	keyData_i	: in  STD_LOGIC_VECTOR (7 downto 0);	-- in, key scan code
	keyClear_o	: out STD_LOGIC;			-- out, scancode has been read
	CharData_o	: out STD_LOGIC_VECTOR (7 downto 0);	-- translated char
	CharAvail_o	: out STD_LOGIC;			-- '1' if translated char is available in char_o
	CharClear_i	: in  STD_LOGIC;			-- '1' to indicate CharData_o has been processed
	StateHistory	: out STD_LOGIC_VECTOR (63 downto 0);
	KeyHistory	: out STD_LOGIC_VECTOR (63 downto 0));
  END COMPONENT;
END PACKAGE;

----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;


entity ps2kbdTranslate is
    PORT (							-- all Signals active high
	clk		: in  STD_LOGIC;
	reset		: in  STD_LOGIC;			-- high active
	keyavail_i	: in  STD_LOGIC;			-- in, new key available
	keyData_i	: in  STD_LOGIC_VECTOR (7 downto 0);	-- in, key scan code
	
	sendData_o	: out STD_LOGIC_VECTOR (7 downto 0);	-- data to be send to keyboard
	sendAvail_o	: out STD_LOGIC;			-- we have data to be send
	sendFinished_i	: in  STD_LOGIC;			-- from ps2 module, '1' if data has been send
	
	keyClear_o	: out STD_LOGIC;			-- out, scancode has been read
	CharData_o	: out STD_LOGIC_VECTOR (7 downto 0);	-- translated char
	CharAvail_o	: out STD_LOGIC;			-- '1' if translated char is available in char_o
	CharClear_i	: in  STD_LOGIC;			-- '1' to indicate CharData_o has been processed
	StateHistory	: out STD_LOGIC_VECTOR (63 downto 0);
	KeyHistory	: out STD_LOGIC_VECTOR (63 downto 0));
end ps2kbdTranslate;

architecture Behavioral of ps2kbdTranslate is


TYPE TKeyTable IS array (0 to 395) OF STD_LOGIC_VECTOR (7 downto 0);
CONSTANT TableShiftOffset : INTEGER := 132;
CONSTANT TableCtrlOffset  : INTEGER := 264;
CONSTANT KeyTable : TKeyTable :=(		-- result in block ram as rom (1 block ram)
-- table w/o shift,ctrl
x"00", -- Scancode x"00"  unused Idx 0
x"B9", -- Scancode x"01"  F9
x"00", -- Scancode x"02"  unused
x"B5", -- Scancode x"03"  F5
x"B3", -- Scancode x"04"  F3
x"B1", -- Scancode x"05"  F1
x"B2", -- Scancode x"06"  F2
x"BC", -- Scancode x"07"  F12
x"00", -- Scancode x"08"  unused
x"BA", -- Scancode x"09"  F10
x"B8", -- Scancode x"0A"  F8
x"B6", -- Scancode x"0B"  F6
x"B4", -- Scancode x"0C"  F4
x"09", -- Scancode x"0D"  TAB
x"5E", -- Scancode x"0E"  ^
x"00", -- Scancode x"0F"  unused
x"00", -- Scancode x"10"  unused
x"00", -- Scancode x"11"  Alt
x"00", -- Scancode x"12"  Shift L
x"00", -- Scancode x"13"  unused
x"00", -- Scancode x"14"  Ctrl L
x"71", -- Scancode x"15"  Q
x"31", -- Scancode x"16"  1!
x"00", -- Scancode x"17"  unused
x"00", -- Scancode x"18"  unused
x"00", -- Scancode x"19"  unused
x"79", -- Scancode x"1A"  Y
x"73", -- Scancode x"1B"  S
x"61", -- Scancode x"1C"  A
x"77", -- Scancode x"1D"  W
x"32", -- Scancode x"1E"  2"
x"00", -- Scancode x"1F"  unused-ext WinL
x"00", -- Scancode x"20"  unused
x"63", -- Scancode x"21"  C
x"78", -- Scancode x"22"  X
x"64", -- Scancode x"23"  D
x"65", -- Scancode x"24"  E
x"34", -- Scancode x"25"  4$
x"33", -- Scancode x"26"  3
x"00", -- Scancode x"27"  unused-ext WinR
x"00", -- Scancode x"28"  unused
x"20", -- Scancode x"29"  Space
x"76", -- Scancode x"2A"  V
x"66", -- Scancode x"2B"  F
x"74", -- Scancode x"2C"  T
x"72", -- Scancode x"2D"  R
x"35", -- Scancode x"2E"  5%
x"00", -- Scancode x"2F"  unused-ext Win2-R
x"00", -- Scancode x"30"  unused
x"6E", -- Scancode x"31"  N
x"62", -- Scancode x"32"  B
x"68", -- Scancode x"33"  H
x"67", -- Scancode x"34"  G
x"7A", -- Scancode x"35"  Z
x"36", -- Scancode x"36"  6&
x"00", -- Scancode x"37"  unused
x"00", -- Scancode x"38"  unused
x"00", -- Scancode x"39"  unused
x"6D", -- Scancode x"3A"  M
x"6A", -- Scancode x"3B"  J
x"75", -- Scancode x"3C"  U
x"37", -- Scancode x"3D"  7/{
x"38", -- Scancode x"3E"  8([
x"00", -- Scancode x"3F"  unused
x"00", -- Scancode x"40"  unused
x"2C", -- Scancode x"41"  ,;
x"6B", -- Scancode x"42"  K
x"69", -- Scancode x"43"  I
x"6F", -- Scancode x"44"  O
x"30", -- Scancode x"45"  0=
x"39", -- Scancode x"46"  9)
x"00", -- Scancode x"47"  unused
x"00", -- Scancode x"48"  unused
x"2E", -- Scancode x"49"  .:
x"2D", -- Scancode x"4A"  -_
x"6C", -- Scancode x"4B"  L
x"7D", -- Scancode x"4C"  
x"70", -- Scancode x"4D"  P
x"7E", -- Scancode x"4E"  ?
x"00", -- Scancode x"4F"  unused
x"00", -- Scancode x"50"  unused
x"00", -- Scancode x"51"  unused
x"7B", -- Scancode x"52"  
x"00", -- Scancode x"53"  unused
x"7C", -- Scancode x"54"  
x"27", -- Scancode x"55"  
x"00", -- Scancode x"56"  unused
x"00", -- Scancode x"57"  unused
x"00", -- Scancode x"58"  CAPS
x"00", -- Scancode x"59"  Shift R
x"0D", -- Scancode x"5A"  ENTER
x"2B", -- Scancode x"5B"  +*~
x"00", -- Scancode x"5C"  unused
x"23", -- Scancode x"5D"  #'
x"00", -- Scancode x"5E"  unused
x"00", -- Scancode x"5F"  unused
x"00", -- Scancode x"60"  unused
x"3C", -- Scancode x"61"  <>|
x"00", -- Scancode x"62"  unused
x"00", -- Scancode x"63"  unused
x"00", -- Scancode x"64"  unused
x"00", -- Scancode x"65"  unused
x"08", -- Scancode x"66"  Backspace
x"00", -- Scancode x"67"  unused
x"00", -- Scancode x"68"  unused
x"31", -- Scancode x"69"  1 NP
x"00", -- Scancode x"6A"  unused
x"34", -- Scancode x"6B"  4 NP
x"37", -- Scancode x"6C"  7 NP
x"00", -- Scancode x"6D"  unused
x"00", -- Scancode x"6E"  unused
x"00", -- Scancode x"6F"  unused
x"30", -- Scancode x"70"  0 NP
x"2E", -- Scancode x"71"  , NP
x"32", -- Scancode x"72"  2 NP
x"35", -- Scancode x"73"  5 NP
x"36", -- Scancode x"74"  6 NP
x"38", -- Scancode x"75"  8 NP
x"1B", -- Scancode x"76"  ESC
x"00", -- Scancode x"77"  Num
x"00", -- Scancode x"78"  F11
x"2B", -- Scancode x"79"  + NP
x"33", -- Scancode x"7A"  3 NP
x"2D", -- Scancode x"7B"  - NP
x"2A", -- Scancode x"7C"  * NP
x"39", -- Scancode x"7D"  9 NP
x"00", -- Scancode x"7E"  ScrollLock
x"00", -- Scancode x"7F"  unused
x"00", -- Scancode x"80"  unused
x"00", -- Scancode x"81"  unused
x"00", -- Scancode x"82"  unused
x"B7", -- Scancode x"83"  F7 Idx 131

       -- Shift Table
x"00", -- Scancode x"00"  unused Idx 132
x"00", -- Scancode x"01"  F9
x"00", -- Scancode x"02"  unused
x"00", -- Scancode x"03"  F5
x"00", -- Scancode x"04"  F3
x"00", -- Scancode x"05"  F1
x"00", -- Scancode x"06"  F2
x"00", -- Scancode x"07"  F12
x"00", -- Scancode x"08"  unused
x"00", -- Scancode x"09"  F10
x"00", -- Scancode x"0A"  F8
x"00", -- Scancode x"0B"  F6
x"00", -- Scancode x"0C"  F4
x"00", -- Scancode x"0D"  TAB
x"5E", -- Scancode x"0E"  ^
x"00", -- Scancode x"0F"  unused
x"00", -- Scancode x"10"  unused
x"00", -- Scancode x"11"  Alt
x"00", -- Scancode x"12"  Shift L
x"00", -- Scancode x"13"  unused
x"00", -- Scancode x"14"  Ctrl L
x"51", -- Scancode x"15"  Q
x"21", -- Scancode x"16"  1!
x"00", -- Scancode x"17"  unused
x"00", -- Scancode x"18"  unused
x"00", -- Scancode x"19"  unused
x"59", -- Scancode x"1A"  Y
x"53", -- Scancode x"1B"  S
x"41", -- Scancode x"1C"  A
x"57", -- Scancode x"1D"  W
x"22", -- Scancode x"1E"  2"
x"00", -- Scancode x"1F"  unused-ext WinL
x"00", -- Scancode x"20"  unused
x"43", -- Scancode x"21"  C
x"58", -- Scancode x"22"  X
x"44", -- Scancode x"23"  D
x"45", -- Scancode x"24"  E
x"24", -- Scancode x"25"  4$
x"40", -- Scancode x"26"  3
x"00", -- Scancode x"27"  unused-ext WinR
x"00", -- Scancode x"28"  unused
x"20", -- Scancode x"29"  Space
x"56", -- Scancode x"2A"  V
x"46", -- Scancode x"2B"  F
x"54", -- Scancode x"2C"  T
x"52", -- Scancode x"2D"  R
x"25", -- Scancode x"2E"  5%
x"00", -- Scancode x"2F"  unused-ext Win2-R
x"00", -- Scancode x"30"  unused
x"4E", -- Scancode x"31"  N
x"42", -- Scancode x"32"  B
x"48", -- Scancode x"33"  H
x"47", -- Scancode x"34"  G
x"5A", -- Scancode x"35"  Z
x"26", -- Scancode x"36"  6&
x"00", -- Scancode x"37"  unused
x"00", -- Scancode x"38"  unused
x"00", -- Scancode x"39"  unused
x"4D", -- Scancode x"3A"  M
x"4A", -- Scancode x"3B"  J
x"55", -- Scancode x"3C"  U
x"2F", -- Scancode x"3D"  7/{
x"28", -- Scancode x"3E"  8([
x"00", -- Scancode x"3F"  unused
x"00", -- Scancode x"40"  unused
x"3B", -- Scancode x"41"  ,;
x"4B", -- Scancode x"42"  K
x"49", -- Scancode x"43"  I
x"4F", -- Scancode x"44"  O
x"3D", -- Scancode x"45"  0=
x"29", -- Scancode x"46"  9)
x"00", -- Scancode x"47"  unused
x"00", -- Scancode x"48"  unused
x"3A", -- Scancode x"49"  .:
x"5F", -- Scancode x"4A"  -_
x"4C", -- Scancode x"4B"  L
x"5D", -- Scancode x"4C"  
x"50", -- Scancode x"4D"  P
x"3F", -- Scancode x"4E"  ?
x"00", -- Scancode x"4F"  unused
x"00", -- Scancode x"50"  unused
x"00", -- Scancode x"51"  unused
x"5B", -- Scancode x"52"  
x"00", -- Scancode x"53"  unused
x"5C", -- Scancode x"54"  
x"60", -- Scancode x"55"  
x"00", -- Scancode x"56"  unused
x"00", -- Scancode x"57"  unused
x"00", -- Scancode x"58"  CAPS
x"00", -- Scancode x"59"  Shift R
x"0A", -- Scancode x"5A"  ENTER
x"2A", -- Scancode x"5B"  +*~
x"00", -- Scancode x"5C"  unused
x"27", -- Scancode x"5D"  #'
x"00", -- Scancode x"5E"  unused
x"00", -- Scancode x"5F"  unused
x"00", -- Scancode x"60"  unused
x"3E", -- Scancode x"61"  <>|
x"00", -- Scancode x"62"  unused
x"00", -- Scancode x"63"  unused
x"00", -- Scancode x"64"  unused
x"00", -- Scancode x"65"  unused
x"00", -- Scancode x"66"  Backspace
x"00", -- Scancode x"67"  unused
x"00", -- Scancode x"68"  unused
x"00", -- Scancode x"69"  1 NP
x"00", -- Scancode x"6A"  unused
x"00", -- Scancode x"6B"  4 NP
x"00", -- Scancode x"6C"  7 NP
x"00", -- Scancode x"6D"  unused
x"00", -- Scancode x"6E"  unused
x"00", -- Scancode x"6F"  unused
x"00", -- Scancode x"70"  0 NP
x"2C", -- Scancode x"71"  , NP
x"00", -- Scancode x"72"  2 NP
x"00", -- Scancode x"73"  5 NP
x"00", -- Scancode x"74"  6 NP
x"00", -- Scancode x"75"  8 NP
x"1B", -- Scancode x"76"  ESC
x"00", -- Scancode x"77"  Num
x"00", -- Scancode x"78"  F11
x"00", -- Scancode x"79"  + NP
x"00", -- Scancode x"7A"  3 NP
x"00", -- Scancode x"7B"  - NP
x"00", -- Scancode x"7C"  * NP
x"00", -- Scancode x"7D"  9 NP
x"00", -- Scancode x"7E"  ScrollLock
x"00", -- Scancode x"7F"  unused
x"00", -- Scancode x"80"  unused
x"00", -- Scancode x"81"  unused
x"00", -- Scancode x"82"  unused
x"00", -- Scancode x"83"  F7 Idx 263

-- ctrl table
x"00", -- Scancode x"00"  unused Idx 264
x"00", -- Scancode x"01"  F9
x"00", -- Scancode x"02"  unused
x"00", -- Scancode x"03"  F5
x"00", -- Scancode x"04"  F3
x"00", -- Scancode x"05"  F1
x"00", -- Scancode x"06"  F2
x"00", -- Scancode x"07"  F12
x"00", -- Scancode x"08"  unused
x"00", -- Scancode x"09"  F10
x"00", -- Scancode x"0A"  F8
x"00", -- Scancode x"0B"  F6
x"00", -- Scancode x"0C"  F4
x"00", -- Scancode x"0D"  TAB
x"1E", -- Scancode x"0E"  ^
x"00", -- Scancode x"0F"  unused
x"00", -- Scancode x"10"  unused
x"00", -- Scancode x"11"  Alt
x"00", -- Scancode x"12"  Shift L
x"00", -- Scancode x"13"  unused
x"00", -- Scancode x"14"  Ctrl L
x"11", -- Scancode x"15"  Q
x"00", -- Scancode x"16"  1!
x"00", -- Scancode x"17"  unused
x"00", -- Scancode x"18"  unused
x"00", -- Scancode x"19"  unused
x"19", -- Scancode x"1A"  Y
x"13", -- Scancode x"1B"  S
x"01", -- Scancode x"1C"  A
x"17", -- Scancode x"1D"  W
x"00", -- Scancode x"1E"  2"
x"00", -- Scancode x"1F"  unused-ext WinL
x"00", -- Scancode x"20"  unused
x"03", -- Scancode x"21"  C
x"18", -- Scancode x"22"  X
x"04", -- Scancode x"23"  D
x"05", -- Scancode x"24"  E
x"00", -- Scancode x"25"  4$
x"00", -- Scancode x"26"  3
x"00", -- Scancode x"27"  unused-ext WinR
x"00", -- Scancode x"28"  unused
x"20", -- Scancode x"29"  Space
x"16", -- Scancode x"2A"  V
x"06", -- Scancode x"2B"  F
x"14", -- Scancode x"2C"  T
x"12", -- Scancode x"2D"  R
x"00", -- Scancode x"2E"  5%
x"00", -- Scancode x"2F"  unused-ext Win2-R
x"00", -- Scancode x"30"  unused
x"0E", -- Scancode x"31"  N
x"02", -- Scancode x"32"  B
x"08", -- Scancode x"33"  H
x"07", -- Scancode x"34"  G
x"1A", -- Scancode x"35"  Z
x"00", -- Scancode x"36"  6&
x"00", -- Scancode x"37"  unused
x"00", -- Scancode x"38"  unused
x"00", -- Scancode x"39"  unused
x"0D", -- Scancode x"3A"  M
x"0A", -- Scancode x"3B"  J
x"15", -- Scancode x"3C"  U
x"00", -- Scancode x"3D"  7/{
x"1B", -- Scancode x"3E"  8([
x"00", -- Scancode x"3F"  unused
x"00", -- Scancode x"40"  unused
x"00", -- Scancode x"41"  ,;
x"0B", -- Scancode x"42"  K
x"09", -- Scancode x"43"  I
x"0F", -- Scancode x"44"  O
x"00", -- Scancode x"45"  0=
x"1D", -- Scancode x"46"  9)
x"00", -- Scancode x"47"  unused
x"00", -- Scancode x"48"  unused
x"00", -- Scancode x"49"  .:
x"1F", -- Scancode x"4A"  -_
x"0C", -- Scancode x"4B"  L
x"00", -- Scancode x"4C"  
x"10", -- Scancode x"4D"  P
x"1C", -- Scancode x"4E"  ?
x"00", -- Scancode x"4F"  unused
x"00", -- Scancode x"50"  unused
x"00", -- Scancode x"51"  unused
x"00", -- Scancode x"52"  
x"00", -- Scancode x"53"  unused
x"00", -- Scancode x"54"  
x"00", -- Scancode x"55"  
x"00", -- Scancode x"56"  unused
x"00", -- Scancode x"57"  unused
x"00", -- Scancode x"58"  CAPS
x"00", -- Scancode x"59"  Shift R
x"0C", -- Scancode x"5A"  ENTER
x"00", -- Scancode x"5B"  +*~
x"00", -- Scancode x"5C"  unused
x"00", -- Scancode x"5D"  #'
x"00", -- Scancode x"5E"  unused
x"00", -- Scancode x"5F"  unused
x"00", -- Scancode x"60"  unused
x"00", -- Scancode x"61"  <>|
x"00", -- Scancode x"62"  unused
x"00", -- Scancode x"63"  unused
x"00", -- Scancode x"64"  unused
x"00", -- Scancode x"65"  unused
x"13", -- Scancode x"66"  Backspace
x"00", -- Scancode x"67"  unused
x"00", -- Scancode x"68"  unused
x"00", -- Scancode x"69"  1 NP
x"00", -- Scancode x"6A"  unused
x"00", -- Scancode x"6B"  4 NP
x"00", -- Scancode x"6C"  7 NP
x"00", -- Scancode x"6D"  unused
x"00", -- Scancode x"6E"  unused
x"00", -- Scancode x"6F"  unused
x"00", -- Scancode x"70"  0 NP
x"00", -- Scancode x"71"  , NP
x"00", -- Scancode x"72"  2 NP
x"00", -- Scancode x"73"  5 NP
x"00", -- Scancode x"74"  6 NP
x"00", -- Scancode x"75"  8 NP
x"1B", -- Scancode x"76"  ESC
x"00", -- Scancode x"77"  Num
x"00", -- Scancode x"78"  F11
x"00", -- Scancode x"79"  + NP
x"00", -- Scancode x"7A"  3 NP
x"00", -- Scancode x"7B"  - NP
x"00", -- Scancode x"7C"  * NP
x"00", -- Scancode x"7D"  9 NP
x"00", -- Scancode x"7E"  ScrollLock
x"00", -- Scancode x"7F"  unused
x"00", -- Scancode x"80"  unused
x"00", -- Scancode x"81"  unused
x"00", -- Scancode x"82"  unused
x"00"  -- Scancode x"83"  F7 Idx 395

);

TYPE TState IS (S_Idle,S_keyAvail_1, S_keyAvail,S_SetLeds,S_SetLeds1,S_SetLeds2,S_SetLeds3,S_SetLeds4,S_CheckExt,
		S_CheckNorm,S_WaitCaps,S_WaitTableData,
		S_WaitCharProcessed,S_WaitResClear);
CONSTANT ALT_KEY	: STD_LOGIC_VECTOR (7 downto 0) := x"11";	-- right alt is E0 11
CONSTANT LEFTSHIFT_KEY	: STD_LOGIC_VECTOR (7 downto 0) := x"12";
CONSTANT RIGHTSHIFT_KEY	: STD_LOGIC_VECTOR (7 downto 0) := x"59";
CONSTANT CTRL_KEY	: STD_LOGIC_VECTOR (7 downto 0) := x"14";	-- right is E0 14
CONSTANT ENTER_KEY	: STD_LOGIC_VECTOR (7 downto 0) := x"5A";	-- right is E0 5A
CONSTANT NUMSLASH_KEY	: STD_LOGIC_VECTOR (7 downto 0) := x"4A";	-- E0 4A, / on numpad
CONSTANT END_KEY	: STD_LOGIC_VECTOR (7 downto 0) := x"69";	-- E0 69, END
CONSTANT HOME_KEY	: STD_LOGIC_VECTOR (7 downto 0) := x"6C";
CONSTANT UP_KEY		: STD_LOGIC_VECTOR (7 downto 0) := x"75";
CONSTANT DOWN_KEY	: STD_LOGIC_VECTOR (7 downto 0) := x"72";
CONSTANT LEFT_KEY	: STD_LOGIC_VECTOR (7 downto 0) := x"6B";
CONSTANT RIGHT_KEY	: STD_LOGIC_VECTOR (7 downto 0) := x"74";
CONSTANT DEL_KEY	: STD_LOGIC_VECTOR (7 downto 0) := x"71";
CONSTANT INS_KEY	: STD_LOGIC_VECTOR (7 downto 0) := x"70";
CONSTANT PGUP_KEY	: STD_LOGIC_VECTOR (7 downto 0) := x"7D";
CONSTANT PGDOWN_KEY	: STD_LOGIC_VECTOR (7 downto 0) := x"7A";

CONSTANT CTRL_Q		: STD_LOGIC_VECTOR (7 downto 0) := x"11";
CONSTANT CTRL_K		: STD_LOGIC_VECTOR (7 downto 0) := x"0B";

CONSTANT SCROLLLOCK_KEY	: STD_LOGIC_VECTOR (7 downto 0) := x"7E";
CONSTANT CAPSLOCK_KEY	: STD_LOGIC_VECTOR (7 downto 0) := x"58";
CONSTANT KeyRelease	: STD_LOGIC_VECTOR (7 downto 0) := x"F0";
CONSTANT KeyExtended	: STD_LOGIC_VECTOR (7 downto 0) := x"E0";

CONSTANT KeybResponseOK : STD_LOGIC_VECTOR (7 downto 0) := x"FA";
CONSTANT KeybResponseERR: STD_LOGIC_VECTOR (7 downto 0) := x"FE";

CONSTANT Stat_ScrollLock: INTEGER := 0;		-- Bit 0 also for LED
CONSTANT Stat_NumLock	: INTEGER := 1;		-- Bit 1 also for LED, currently not used
CONSTANT Stat_CapsLock	: INTEGER := 2;		-- Bit 2 also for LED
CONSTANT Stat_Shift	: INTEGER := 3;		-- shift pressed (internal)
CONSTANT Stat_Alt	: INTEGER := 4;		-- left or right alt (internal)
CONSTANT Stat_Ctrl	: INTEGER := 5;		-- left or right ctrl (internal)
SIGNAL ShiftStatus	: STD_LOGIC_VECTOR (5 downto 0);

SIGNAL ScanCode		: STD_LOGIC_VECTOR (7 downto 0);
SIGNAL State		: TState;
SIGNAL isExtended	: BOOLEAN;				-- extended scan code
SIGNAL isKeyRelease	: BOOLEAN;
--SIGNAL SkipNext		: BOOLEAN;
SIGNAL CharData		: STD_LOGIC_VECTOR (7 downto 0);
SIGNAL SecondCharData	: STD_LOGIC_VECTOR (7 downto 0);
SIGNAL SecondChar	: STD_LOGIC;				-- second char for a key to send
SIGNAL StateHist	: STD_LOGIC_VECTOR (63 downto 0);
SIGNAL KeyHist		: STD_LOGIC_VECTOR (63 downto 0);
SIGNAL SendWaitCnt	: INTEGER RANGE 0 TO 255;

function State2Logic (st:TState) return STD_LOGIC_VECTOR is
  VARIABLE result : std_logic_vector (7 downto 0);
  begin
    CASE State IS
      WHEN S_Idle		=> Result := x"01";
      WHEN S_keyAvail_1		=> Result := x"02";
      WHEN S_keyAvail		=> Result := x"03";
      WHEN S_SetLeds		=> Result := x"04";
      WHEN S_SetLeds1		=> Result := x"05";
      WHEN S_SetLeds2		=> Result := x"06";
      WHEN S_SetLeds3		=> Result := x"07";
      WHEN S_SetLeds4		=> Result := x"08";
      WHEN S_CheckExt		=> Result := x"09";
      WHEN S_CheckNorm		=> Result := x"0A";
      WHEN S_WaitCaps		=> Result := x"0B";
      WHEN S_WaitTableData	=> Result := x"0C";
      WHEN S_WaitCharProcessed	=> Result := x"0D";
      WHEN S_WaitResClear	=> Result := x"0E";
      WHEN Others		=> Result := x"FF";
    END CASE;
    return result;
  end State2Logic;

begin
  StateHistory <= StateHist;
  KeyHistory <= KeyHist;
  
  x:PROCESS(clk,reset)
  VARIABLE StateDebug : STD_LOGIC_VECTOR(7 downto 0);
  VARIABLE KeyDebug : STD_LOGIC_VECTOR(7 downto 0);
  BEGIN
    IF Reset = '1' THEN
      ScanCode <= (Others => '0');
      State <= S_Idle;
      isExtended <= false;
      isKeyRelease <= false;
      ShiftStatus <= (Others => '0');
      keyClear_o <= '0';
      CharAvail_o <= '0';
      SecondChar <= '0';
      sendAvail_o <= '0';
      StateHist  <= (Others => '0');
      KeyHist  <= (Others => '0');
    ELSIF Rising_Edge(Clk) THEN
      
      StateDebug := State2Logic (State);
      IF StateHist(7 downto 0) /= StateDebug THEN	-- to allow reading state history from host
        StateHist <= StateHist (55 downto 0) & StateDebug;
      END IF;
    
      CASE State IS
        WHEN S_Idle =>		keyClear_o <= '0';
				IF SecondChar = '1' THEN	-- second char of 2 byte key
				  SecondChar <= '0';
				  CharData <= SecondCharData;
				  State <= S_WaitTableData;
				ELSIF keyavail_i = '1' THEN	-- if new scancode available
				  ScanCode <= keyData_i;	-- save it
				  State <= S_keyAvail_1;
				  keyClear_o <= '1';		-- scancode has been saved to ps2 module
				  KeyHist <= KeyHist (55 downto 0) & keyData_i;
				END IF;
	WHEN S_keyAvail_1 =>	State <= S_keyAvail;
	WHEN S_keyAvail =>	IF isKeyRelease THEN
				  isKeyRelease <= false;
				  isExtended <= false;
				  CASE ScanCode IS
				    WHEN ALT_KEY	=> ShiftStatus(Stat_Alt) <= '0';
				    WHEN LEFTSHIFT_KEY	=> ShiftStatus(Stat_Shift) <= '0';
				    WHEN RIGHTSHIFT_KEY	=> ShiftStatus(Stat_Shift) <= '0';
				    WHEN CTRL_KEY	=> ShiftStatus(Stat_Ctrl) <= '0';
				    WHEN Others 	=> NULL;
				  END CASE;
				  State <= S_Idle;
				ELSIF ScanCode = KeyRelease THEN
				  isKeyRelease <= true;
				  State <= S_Idle;
				ELSIF ScanCode = KeyExtended THEN
				  isExtended <= true;
				  State <= S_Idle;
				ELSE
				  CASE ScanCode IS
				    WHEN KeybResponseOK => State <= S_Idle;	-- response from
				    WHEN KeybResponseERR=> State <= S_Idle;	-- set leds
				    WHEN ALT_KEY	=> ShiftStatus(Stat_Alt) <= '1';
							   State <= S_Idle;
				    WHEN LEFTSHIFT_KEY	=> ShiftStatus(Stat_Shift) <= '1';
							   State <= S_Idle;
				    WHEN RIGHTSHIFT_KEY	=> ShiftStatus(Stat_Shift) <= '1';
							   State <= S_Idle;
				    WHEN CTRL_KEY	=> ShiftStatus(Stat_Ctrl) <= '1';
							   State <= S_Idle;
							   -- TODO: what about repeat for caps/scroll lock
				    WHEN SCROLLLOCK_KEY	=> ShiftStatus(Stat_ScrollLock) <= NOT ShiftStatus(Stat_ScrollLock);
							   State <= S_SetLeds;
				    WHEN CAPSLOCK_KEY	=> ShiftStatus(Stat_CapsLock) <= NOT ShiftStatus(Stat_CapsLock);
							   State <= S_SetLeds;
				    WHEN Others 	=> IF isExtended THEN
							     state <= S_CheckExt;
							   ELSE
							     State <= S_CheckNorm;
							   END IF;
				  END CASE;
				END IF;
	WHEN S_setLeds =>	-- TODO: does not work yet ????
				keyClear_o <= '0';
				sendData_o <= x"ED";		-- set LED command
				State <= S_setLeds1;
	WHEN S_SetLeds1 =>	sendAvail_o <= '1';		-- we have data to be send
				State <= S_SetLeds2;
	WHEN S_SetLeds2 =>	sendAvail_o <= '0';
				IF sendFinished_i = '1' THEN
				  State <= S_setLeds3;
				  SendWaitCnt <= 255;
				  sendData_o <= "00000" & ShiftStatus(2 downto 0);
				END IF;
	WHEN S_SetLeds3 =>	SendWaitCnt <= SendWaitCnt - 1;
				IF SendWaitCnt = 0 THEN
				  State <= S_SetLeds4;
				  sendAvail_o <= '1';		-- we have data to be send
				END IF;
	WHEN S_SetLeds4 =>	sendAvail_o <= '0';
				IF sendFinished_i = '1' THEN
				  State <= S_Idle;
				END IF;
	WHEN S_CheckExt =>	-- check extended keys
				isExtended <= false;
				CASE ScanCode IS
				  WHEN ALT_KEY =>	State <= S_keyAvail;	-- for right alt key
				  WHEN CTRL_KEY =>	State <= S_keyAvail;	-- for right ctrl key
				  WHEN ENTER_KEY =>	State <= S_keyAvail;	-- Enter on numpad
				  WHEN NUMSLASH_KEY =>	CharData <= x"2F";	-- set /
							State <= S_WaitTableData;
				  WHEN END_KEY =>	IF ShiftStatus(Stat_Ctrl) = '0' THEN
							  SecondChar <= '1';	-- send ^Q D - end of line
							  CharData <= Ctrl_Q;
							  SecondCharData <= x"44";
							ELSE
							  SecondChar <= '1';	-- send ^Q C - end of file
							  CharData <= Ctrl_Q;
							  SecondCharData <= x"43";
							END IF;
							State <= S_WaitTableData;
				  WHEN HOME_KEY =>	IF ShiftStatus(Stat_Ctrl) = '0' THEN
							  SecondChar <= '1';	-- send ^Q S - line begin
							  CharData <= Ctrl_Q;
							  SecondCharData <= x"53";
							ELSE
							  SecondChar <= '1';	-- send ^Q R - file begin
							  CharData <= Ctrl_Q;
							  SecondCharData <= x"52";
							END IF;
							State <= S_WaitTableData;
				  WHEN UP_KEY =>	CharData <= x"05";	-- ^E
							State <= S_WaitTableData;
				  WHEN DOWN_KEY =>	CharData <= x"18";	-- ^X
							State <= S_WaitTableData;
				  WHEN LEFT_KEY =>	IF ShiftStatus(Stat_Ctrl) = '0' THEN
							  CharData <= x"13";	-- ^S
							ELSE
							  CharData <= x"01";	-- ^A
							END IF;
							State <= S_WaitTableData;
				  WHEN RIGHT_KEY =>	IF ShiftStatus(Stat_Ctrl) = '0' THEN
							  CharData <= x"04";	-- ^D
							ELSE
							  CharData <= x"06";	-- ^F
							END IF;
							State <= S_WaitTableData;
				  WHEN DEL_KEY =>	CharData <= x"7F";
							State <= S_WaitTableData;
				  WHEN INS_KEY =>	CharData <= x"16";	-- ^V
							State <= S_WaitTableData;
				  WHEN PGUP_KEY =>	IF ShiftStatus(Stat_Ctrl) = '0' THEN
							  CharData <= x"12";	-- ^R
							ELSE
							  SecondChar <= '1';	-- send ^Q E
							  CharData <= Ctrl_Q;
							  SecondCharData <= x"45";
							END IF;
							State <= S_WaitTableData;
				  WHEN PGDOWN_KEY =>	IF ShiftStatus(Stat_Ctrl) = '0' THEN
							  CharData <= x"03";	-- ^C
							ELSE
							  SecondChar <= '1';	-- send ^Q X
							  CharData <= Ctrl_Q;
							  SecondCharData <= x"58";
							END IF;
							State <= S_WaitTableData;
				  WHEN Others =>	State <= S_Idle;
				END CASE;
	WHEN S_CheckNorm =>	IF ShiftStatus(Stat_Alt) = '1' THEN
				  CASE ScanCode IS
				    WHEN x"15"  =>	CharData <= x"40"; -- ALT Q -> @
							State <= S_WaitTableData;
				    WHEN x"3D"	=>	CharData <= x"7B"; -- ALT 7 -> {
							State <= S_WaitTableData;
				    WHEN x"3E"	=>	CharData <= x"5B"; -- ALT 8 -> [
							State <= S_WaitTableData;
				    WHEN x"45"	=>	CharData <= x"7D"; -- ALT 0 -> }
							State <= S_WaitTableData;
				    WHEN x"46"	=>	CharData <= x"5D"; -- ALT 9 -> ]
							State <= S_WaitTableData;
				    WHEN x"4E"	=>	CharData <= x"5C"; -- ALT ? -> \
							State <= S_WaitTableData;
				    WHEN x"5B"	=>	CharData <= x"7E"; -- ALT +* -> ~
							State <= S_WaitTableData;
				    WHEN x"61"	=>	CharData <= x"7C"; -- ALT <> -> |
							State <= S_WaitTableData;
				    WHEN others =>	State <= S_Idle;
				  END CASE;
				ELSIF ShiftStatus(Stat_Ctrl) = '1' THEN
				-- TODO: check bounds
				  CharData <= KeyTable(conv_integer(ScanCode)+TableCtrlOffset);
				ELSIF ShiftStatus(Stat_Shift) = '1' THEN
				  -- TODO: check bounds
				  CharData <= KeyTable(conv_integer(ScanCode)+TableShiftOffset);
				ELSE
				  -- TODO: check bounds
				  CharData <= KeyTable(conv_integer(ScanCode));
				END IF;
				State <= S_WaitTableData;
	WHEN S_WaitTableData =>	-- check if capslock is on add change lower in upper case if on
				IF ShiftStatus(Stat_CapsLock) = '1' THEN
				  IF (CharData >= x"61") and (CharData <= x"7A") THEN
				    CharData <= CharData -x"20";
				    -- TODO: German umlauts mussing here
				  END IF;
				END IF;
				State <= S_WaitCaps;
	WHEN S_WaitCaps =>	CharData_o <= CharData;		-- set output data
				State <= S_WaitCharProcessed;
	WHEN S_WaitCharProcessed =>
				CharAvail_o <= '1';
				IF CharClear_i = '1' THEN	-- has been processed
				  State <= S_WaitResClear;
				END IF;
	WHEN S_WaitResClear =>	CharAvail_o <= '0';
				IF CharClear_i = '0' THEN	-- and wait for zero
				  State <= S_Idle;
				END IF;
      END CASE;
    END IF;
  END PROCESS;

end Behavioral;

