----------------------------------------------------------------------------------
-- Engineer:       Armin Diehl <diehl@nordrhein.de>
-- Create Date:    23:10:51 12/05/2006 
-- Design Name:    clockgen
-- Module Name:    vga_clockgen - Behavioral 
-- Project Name:   vga textmode video
-- Description:    generates clock for vga textmode
-- Revision:       0.01 - initial version
----------------------------------------------------------------------------------

--global Buf Clock2_bufg : bufg port map(i => clock2, o => clock2_int);

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

PACKAGE vga_clockgen_pckg IS
  COMPONENT vga_clockgen IS
    PORT ( reset          	: IN   STD_LOGIC;	-- global reset (HIGH active)
	   clk            	: IN   STD_LOGIC;	-- 100 MHz
           clk50M         	: OUT  STD_LOGIC;	-- 50 MHz out
           clk25M         	: OUT  STD_LOGIC;	-- 25 MHz out
           clkCursorBlink 	: OUT  STD_LOGIC;	-- clk for blinking cursor
	   clkCursorFastBlink	: OUT  STD_LOGIC;
           clkCharBlink   	: OUT  STD_LOGIC);	-- clk for blinking characters
  END COMPONENT;
END PACKAGE;

----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

ENTITY vga_clockgen IS
    PORT ( reset          	: IN   STD_LOGIC;	-- global reset (HIGH active)
	   clk            	: IN   STD_LOGIC;	-- 100 MHz
           clk50M         	: OUT  STD_LOGIC;	-- 50 MHz out
           clk25M         	: OUT  STD_LOGIC;	-- 25 MHz out
           clkCursorBlink 	: OUT  STD_LOGIC;	-- clk for blinking cursor
	   clkCursorFastBlink	: OUT  STD_LOGIC;
           clkCharBlink   	: OUT  STD_LOGIC);	-- clk for blinking characters
END vga_clockgen;


ARCHITECTURE Behavioral OF vga_clockgen IS
 
 SIGNAL clk25Mi         : STD_LOGIC;
 SIGNAL clk50Mi         : STD_LOGIC;
 ATTRIBUTE buffer_type: STRING;
 ATTRIBUTE buffer_type OF clk25M: SIGNAL IS "bufg";
 ATTRIBUTE buffer_type OF clk50M: SIGNAL IS "bufg";
 SIGNAL counter		: STD_LOGIC_VECTOR(24 downto 0);
 SIGNAL cnt		: STD_LOGIC_VECTOR(1 downto 0);
 

BEGIN

-- generate 50 and 25 MHz clocks
  genclks:PROCESS (reset,clk)
  BEGIN
    IF Rising_Edge(clk) THEN
      IF Reset = '1' THEN
        cnt <= "00";
      ELSE
        cnt <= cnt + 1;
      END IF;
      clk50Mi <= cnt(0);
      clk25Mi <= cnt(1);
    END IF;
  END PROCESS;
  

-- generate clocks for cursor blink and character blink  
  genBlink:PROCESS(reset,clk25Mi)
  BEGIN
    IF reset = '1' THEN
      counter <= (others => '0');
    ELSIF Rising_Edge(clk25Mi) THEN
      Counter <= Counter + 1;
    END IF;
  END PROCESS;
  
  clk50M <= clk50Mi;
  clk25M <= clk25Mi;
  clkCursorFastBlink <= counter(22);
  clkCursorBlink <= counter(23);
  clkCharBlink <= counter(24); 
END Behavioral;

